import base64
import io
import json
import logging

import numpy as np
import pytest
from google.protobuf import json_format
from PIL import Image

from seldon_core.flask_utils import SeldonMicroserviceException
from seldon_core.imports_helper import _TF_PRESENT
from seldon_core.metrics import SeldonMetrics
from seldon_core.proto import prediction_pb2
from seldon_core.user_model import SeldonComponent
from seldon_core.utils import json_to_seldon_message, seldon_message_to_json
from seldon_core.wrapper import SeldonModelGRPC, get_grpc_server, get_rest_microservice

from .utils import skipif_tf_missing

if _TF_PRESENT:
    import tensorflow as tf
    from tensorflow.core.framework.tensor_pb2 import TensorProto

HEALTH_PING_URL = "/health/ping"
HEALTH_STATUS_URL = "/health/status"
METADATA_URL = "/metadata"

"""
 Checksum of bytes. Used to check data integrity of binData passed in multipart/form-data request

 Parameters
 ----------
  the_bytes
    Input bytes

  Returns
  -------
  the checksum
"""


def rs232_checksum(the_bytes):
    return b"%02X" % (sum(the_bytes) & 0xFF)


class UserObject(SeldonComponent):
    HEALTH_STATUS_RESPONSE = [0.123]
    METADATA_RESPONSE = {
        "name": "my-model-name",
        "versions": ["model-version"],
        "platform": "model-platform",
        "inputs": [{"name": "input", "datatype": "BYTES", "shape": [1]}],
        "outputs": [{"name": "output", "datatype": "BYTES", "shape": [1]}],
        "custom": {},
    }

    def __init__(self, metrics_ok=True, ret_nparray=False, ret_meta=False):
        self.metrics_ok = metrics_ok
        self.ret_nparray = ret_nparray
        self.nparray = np.array([1, 2, 3])
        self.ret_meta = ret_meta

        self.rewards = []

    def predict(self, X, features_names, **kwargs):
        """
        Return a prediction.

        Parameters
        ----------
        X : array-like
        feature_names : array of feature names (optional)
        """
        if self.ret_meta:
            self.inc_meta = kwargs.get("meta")
        if self.ret_nparray:
            return self.nparray
        else:
            logging.info("Predict called - will run identity function")
            logging.info(X)
            return X

    def send_feedback(self, features, feature_names, reward, truth, routing=None):
        self.rewards.append(reward)
        logging.info("Feedback called")

    def tags(self):
        if self.ret_meta:
            return {"inc_meta": self.inc_meta}
        else:
            return {"mytag": 1}

    def metrics(self):
        if self.metrics_ok:
            return [{"type": "COUNTER", "key": "mycounter", "value": 1}]
        else:
            return [{"type": "BAD", "key": "mycounter", "value": 1}]

    def health_status(self):
        return self.predict(self.HEALTH_STATUS_RESPONSE, ["some_float"])

    def init_metadata(self):
        return self.METADATA_RESPONSE


class UserObjectLowLevel(SeldonComponent):
    HEALTH_STATUS_RAW_RESPONSE = [123.456, 7.89]

    def __init__(self, metrics_ok=True, ret_nparray=False):
        self.metrics_ok = metrics_ok
        self.ret_nparray = ret_nparray
        self.nparray = np.array([1, 2, 3])

    def predict_rest(self, request):
        return {"data": {"ndarray": [9, 9]}}

    def predict_grpc(self, request):
        arr = np.array([9, 9])
        datadef = prediction_pb2.DefaultData(
            tensor=prediction_pb2.Tensor(shape=(2, 1), values=arr)
        )
        request = prediction_pb2.SeldonMessage(data=datadef)
        return request

    def send_feedback_rest(self, request):
        logging.info("Feedback called")

    def send_feedback_grpc(self, request):
        logging.info("Feedback called")

    def health_status_raw(self):
        return {"data": {"ndarray": self.HEALTH_STATUS_RAW_RESPONSE}}


class UserObjectLowLevelWithStatusInResponse(SeldonComponent):
    def __init__(self, metrics_ok=True, ret_nparray=False):
        self.metrics_ok = metrics_ok
        self.ret_nparray = ret_nparray
        self.nparray = np.array([1, 2, 3])

    def predict_rest(self, request):
        return {
            "data": {"ndarray": [9, 9]},
            "status": {"code": 400, "status": "FAILURE"},
        }

    def predict_grpc(self, request):
        arr = np.array([9, 9])
        datadef = prediction_pb2.DefaultData(
            tensor=prediction_pb2.Tensor(shape=(2, 1), values=arr)
        )
        request = prediction_pb2.SeldonMessage(data=datadef)
        return request

    def send_feedback_rest(self, request):
        logging.info("Feedback called")

    def send_feedback_grpc(self, request):
        logging.info("Feedback called")


class UserObjectLowLevelWithStatusInResponseWithPredictRaw(SeldonComponent):
    def __init__(self, check_name):
        self.check_name = check_name

    def predict_raw(self, msg):
        msg = json_to_seldon_message(msg)
        if self.check_name == "img":
            file_data = msg.binData
            img = Image.open(io.BytesIO(file_data))
            img.verify()
            return {
                "meta": seldon_message_to_json(msg.meta),
                "data": {"ndarray": [rs232_checksum(file_data).decode("utf-8")]},
                "status": {"code": 400, "status": "FAILURE"},
            }
        elif self.check_name == "txt":
            file_data = msg.binData
            return {
                "meta": seldon_message_to_json(msg.meta),
                "data": {"ndarray": [file_data.decode("utf-8")]},
                "status": {"code": 400, "status": "FAILURE"},
            }
        elif self.check_name == "strData":
            file_data = msg.strData
            return {
                "meta": seldon_message_to_json(msg.meta),
                "data": {"ndarray": [file_data]},
                "status": {"code": 400, "status": "FAILURE"},
            }


class UserObjectLowLevelWithPredictRaw(SeldonComponent):
    def __init__(self, check_name):
        self.check_name = check_name

    def predict_raw(self, msg):
        msg = json_to_seldon_message(msg)
        if self.check_name == "img":
            file_data = msg.binData
            img = Image.open(io.BytesIO(file_data))
            img.verify()
            return {
                "meta": seldon_message_to_json(msg.meta),
                "data": {"ndarray": [rs232_checksum(file_data).decode("utf-8")]},
            }
        elif self.check_name == "txt":
            file_data = msg.binData
            return {
                "meta": seldon_message_to_json(msg.meta),
                "data": {"ndarray": [file_data.decode("utf-8")]},
            }
        elif self.check_name == "strData":
            file_data = msg.strData
            return {
                "meta": seldon_message_to_json(msg.meta),
                "data": {"ndarray": [file_data]},
            }


class UserObjectLowLevelGrpc(SeldonComponent):
    def __init__(self, metrics_ok=True, ret_nparray=False):
        self.metrics_ok = metrics_ok
        self.ret_nparray = ret_nparray
        self.nparray = np.array([1, 2, 3])

    def predict_grpc(self, request):
        arr = np.array([9, 9])
        datadef = prediction_pb2.DefaultData(
            tensor=prediction_pb2.Tensor(shape=(2, 1), values=arr)
        )
        request = prediction_pb2.SeldonMessage(data=datadef)
        return request

    def send_feedback_rest(self, request):
        logging.info("Feedback called")

    def send_feedback_grpc(self, request):
        logging.info("Feedback called")


@pytest.fixture(name="mock_get_model_name")
def fixture_get_model_name(mocker):
    return mocker.patch(
        "seldon_core.utils.get_model_name", autospec=True, return_value="my-test-model"
    )


@pytest.fixture(name="mock_get_image_name")
def fixture_get_image_name(mocker):
    return mocker.patch(
        "seldon_core.utils.get_image_name",
        autospec=True,
        return_value="my-test-model-image",
    )


def test_model_ok():
    user_object = UserObject()
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()
    rv = client.get('/predict?json={"data":{"names":["a","b"],"ndarray":[[1,2]]}}')
    j = json.loads(rv.data)
    logging.info(j)
    assert rv.status_code == 200
    assert j["meta"]["tags"] == {"mytag": 1}
    assert j["meta"]["metrics"][0]["key"] == user_object.metrics()[0]["key"]
    assert j["meta"]["metrics"][0]["value"] == user_object.metrics()[0]["value"]
    assert j["data"]["names"] == ["t:0", "t:1"]
    assert j["data"]["ndarray"] == [[1.0, 2.0]]


def test_model_v01_ok():
    user_object = UserObject()
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()

    payload = {"data": {"names": ["a", "b"], "ndarray": [[1, 2]]}}
    rv = client.post("/api/v0.1/predictions", json=payload)
    j = json.loads(rv.data)
    logging.info(j)
    assert rv.status_code == 200
    assert j["meta"]["tags"] == {"mytag": 1}
    assert j["meta"]["metrics"][0]["key"] == user_object.metrics()[0]["key"]
    assert j["meta"]["metrics"][0]["value"] == user_object.metrics()[0]["value"]
    assert j["data"]["names"] == ["t:0", "t:1"]
    assert j["data"]["ndarray"] == [[1.0, 2.0]]


def test_model_v10_ok():
    user_object = UserObject()
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()

    payload = {"data": {"names": ["a", "b"], "ndarray": [[1, 2]]}}
    rv = client.post("/api/v1.0/predictions", json=payload)
    j = json.loads(rv.data)
    logging.info(j)
    assert rv.status_code == 200
    assert j["meta"]["tags"] == {"mytag": 1}
    assert j["meta"]["metrics"][0]["key"] == user_object.metrics()[0]["key"]
    assert j["meta"]["metrics"][0]["value"] == user_object.metrics()[0]["value"]
    assert j["data"]["names"] == ["t:0", "t:1"]
    assert j["data"]["ndarray"] == [[1.0, 2.0]]


def test_model_puid_ok():
    user_object = UserObject()
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()
    rv = client.get(
        '/predict?json={"meta":{"puid":"123"},"data":{"names":["a","b"],"ndarray":[[1,2]]}}'
    )
    j = json.loads(rv.data)
    logging.info(j)
    assert rv.status_code == 200
    assert j["meta"]["tags"] == {"mytag": 1}
    assert j["meta"]["metrics"][0]["key"] == user_object.metrics()[0]["key"]
    assert j["meta"]["metrics"][0]["value"] == user_object.metrics()[0]["value"]
    assert j["data"]["names"] == ["t:0", "t:1"]
    assert j["data"]["ndarray"] == [[1.0, 2.0]]
    assert j["meta"]["puid"] == "123"


def test_requestPath_ok(mock_get_model_name, mock_get_image_name):
    user_object = UserObject()
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()
    rv = client.get(
        '/predict?json={"meta":{"puid":"123"},"data":{"names":["a","b"],"ndarray":[[1,2]]}}'
    )
    j = json.loads(rv.data)
    logging.info(j)
    assert rv.status_code == 200
    assert j["meta"]["requestPath"] == {"my-test-model": "my-test-model-image"}


def test_requestPath_2nd_node_ok(mock_get_model_name, mock_get_image_name):
    user_object = UserObject()
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()
    rv = client.get(
        '/predict?json={"meta":{"requestPath":{"earlier-node": "earlier-image"}},"data":{"names":["a","b"],"ndarray":[[1,2]]}}'
    )
    j = json.loads(rv.data)
    logging.info(j)
    assert rv.status_code == 200
    assert j["meta"]["requestPath"] == {
        "my-test-model": "my-test-model-image",
        "earlier-node": "earlier-image",
    }


def test_proto_requestPath_ok(mock_get_model_name, mock_get_image_name):
    user_object = UserObject()
    seldon_metrics = SeldonMetrics()
    app = SeldonModelGRPC(user_object, seldon_metrics)
    arr = np.array([1, 2])
    datadef = prediction_pb2.DefaultData(
        tensor=prediction_pb2.Tensor(shape=(2, 1), values=arr)
    )
    meta = prediction_pb2.Meta()
    json_format.ParseDict({"tags": {"foo": "bar"}}, meta)
    request = prediction_pb2.SeldonMessage(data=datadef, meta=meta)
    resp = app.Predict(request, None)
    jStr = json_format.MessageToJson(resp)
    j = json.loads(jStr)
    logging.info(j)
    assert j["meta"]["requestPath"] == {"my-test-model": "my-test-model-image"}


def test_proto_requestPath_2nd_node_ok(mock_get_model_name, mock_get_image_name):
    user_object = UserObject()
    seldon_metrics = SeldonMetrics()
    app = SeldonModelGRPC(user_object, seldon_metrics)
    arr = np.array([1, 2])
    datadef = prediction_pb2.DefaultData(
        tensor=prediction_pb2.Tensor(shape=(2, 1), values=arr)
    )
    meta = prediction_pb2.Meta()
    json_format.ParseDict({"requestPath": {"earlier-node": "earlier-image"}}, meta)
    request = prediction_pb2.SeldonMessage(data=datadef, meta=meta)
    resp = app.Predict(request, None)
    jStr = json_format.MessageToJson(resp)
    j = json.loads(jStr)
    logging.info(j)
    assert j["meta"]["requestPath"] == {
        "my-test-model": "my-test-model-image",
        "earlier-node": "earlier-image",
    }


def test_model_lowlevel_ok():
    user_object = UserObjectLowLevel()
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()
    rv = client.get('/predict?json={"data":{"ndarray":[1,2]}}')
    j = json.loads(rv.data)
    logging.info(j)
    assert rv.status_code == 200
    assert j["data"]["ndarray"] == [9, 9]


def test_model_lowlevel_multi_form_data_text_file_ok():
    user_object = UserObjectLowLevelWithPredictRaw("txt")
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()
    rv = client.post(
        "/predict",
        data={
            "meta": '{"puid":"1234"}',
            "binData": (f"./tests/resources/test.txt", "test.txt"),
        },
        content_type="multipart/form-data",
    )
    j = json.loads(rv.data)
    assert rv.status_code == 200
    assert j["meta"]["puid"] == "1234"
    assert (
        j["data"]["ndarray"][0]
        == "this is test file for testing multipart/form-data input\n"
    )


def test_model_lowlevel_multi_form_data_img_file_ok():
    user_object = UserObjectLowLevelWithPredictRaw("img")
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()
    rv = client.post(
        "/predict",
        data={
            "meta": '{"puid":"1234"}',
            "binData": (f"./tests/resources/test.png", "test.png"),
        },
        content_type="multipart/form-data",
    )
    j = json.loads(rv.data)
    assert rv.status_code == 200
    assert j["meta"]["puid"] == "1234"
    with open("./tests/resources/test.png", "rb") as f:
        img_data = f.read()
    assert j["data"]["ndarray"][0] == rs232_checksum(img_data).decode("utf-8")


def test_model_lowlevel_multi_form_data_strData_ok():
    user_object = UserObjectLowLevelWithPredictRaw("strData")
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()
    rv = client.post(
        "/predict",
        data={
            "meta": '{"puid":"1234"}',
            "strData": (f"./tests/resources/test.txt", "test.txt"),
        },
        content_type="multipart/form-data",
    )
    j = json.loads(rv.data)
    assert rv.status_code == 200
    assert j["meta"]["puid"] == "1234"
    assert (
        j["data"]["ndarray"][0]
        == "this is test file for testing multipart/form-data input\n"
    )


def test_model_lowlevel_multi_form_data_strData_non200status():
    user_object = UserObjectLowLevelWithStatusInResponseWithPredictRaw("strData")
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()
    rv = client.post(
        "/predict",
        data={
            "meta": '{"puid":"1234"}',
            "strData": (f"./tests/resources/test.txt", "test.txt"),
        },
        content_type="multipart/form-data",
    )
    j = json.loads(rv.data)
    assert rv.status_code == 400
    assert j["meta"]["puid"] == "1234"
    assert (
        j["data"]["ndarray"][0]
        == "this is test file for testing multipart/form-data input\n"
    )


def test_model_multi_form_data_ok():
    user_object = UserObject()
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()
    rv = client.post(
        "/predict",
        data={"data": '{"names":["a","b"],"ndarray":[[1,2]]}'},
        content_type="multipart/form-data",
    )
    j = json.loads(rv.data)
    logging.info(j)
    assert rv.status_code == 200
    assert j["meta"]["tags"] == {"mytag": 1}
    assert j["meta"]["metrics"][0]["key"] == user_object.metrics()[0]["key"]
    assert j["meta"]["metrics"][0]["value"] == user_object.metrics()[0]["value"]
    assert j["data"]["names"] == ["t:0", "t:1"]
    assert j["data"]["ndarray"] == [[1.0, 2.0]]


def test_model_feedback_ok():
    user_object = UserObject()
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()
    rv = client.get(
        '/send-feedback?json={"request":{"data":{"ndarray":[]}},"reward":1.0}'
    )
    j = json.loads(rv.data)
    logging.info(j)
    assert rv.status_code == 200
    assert user_object.rewards[-1] == 1.0


def test_feedback_v10_ok():
    user_object = UserObject()
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()

    payload = {
        "request": {"data": {"names": ["a", "b"], "ndarray": [[1, 2]]}},
        "reward": 1.0,
    }
    rv = client.post("/api/v1.0/feedback", json=payload)
    j = json.loads(rv.data)
    logging.info(j)
    assert rv.status_code == 200


def test_feedback_v01_ok():
    user_object = UserObject()
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()

    payload = {
        "request": {"data": {"names": ["a", "b"], "ndarray": [[1, 2]]}},
        "reward": 1.0,
    }
    rv = client.post("/api/v0.1/feedback", json=payload)
    j = json.loads(rv.data)
    logging.info(j)
    assert rv.status_code == 200


def test_model_feedback_lowlevel_ok():
    user_object = UserObjectLowLevel()
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()
    rv = client.get(
        '/send-feedback?json={"request":{"data":{"ndarray":[]}},"reward":1.0}'
    )
    j = json.loads(rv.data)
    logging.info(j)
    assert rv.status_code == 200


def test_model_non200status_lowlevel():
    user_object = UserObjectLowLevelWithStatusInResponse()
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()
    rv = client.get('/predict?json={"request":{"data":{"ndarray":[]}},"reward":1.0}')
    j = json.loads(rv.data)
    logging.info(j)
    assert rv.status_code == 400


@skipif_tf_missing
def test_model_tftensor_ok():
    user_object = UserObject()
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()
    arr = np.array([1, 2])
    datadef = prediction_pb2.DefaultData(tftensor=tf.make_tensor_proto(arr))
    request = prediction_pb2.SeldonMessage(data=datadef)
    jStr = json_format.MessageToJson(request)
    rv = client.get("/predict?json=" + jStr)
    j = json.loads(rv.data)
    logging.info(j)
    assert rv.status_code == 200
    assert j["meta"]["tags"] == {"mytag": 1}
    assert j["meta"]["metrics"][0]["key"] == user_object.metrics()[0]["key"]
    assert j["meta"]["metrics"][0]["value"] == user_object.metrics()[0]["value"]
    assert "tftensor" in j["data"]
    tfp = TensorProto()
    json_format.ParseDict(j["data"].get("tftensor"), tfp, ignore_unknown_fields=False)
    arr2 = tf.make_ndarray(tfp)
    assert np.array_equal(arr, arr2)


def test_model_ok_with_names():
    user_object = UserObject()
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()
    rv = client.get('/predict?json={"data":{"names":["a","b"],"ndarray":[[1,2]]}}')
    j = json.loads(rv.data)
    assert rv.status_code == 200
    assert j["meta"]["tags"] == {"mytag": 1}
    assert j["meta"]["metrics"][0]["key"] == user_object.metrics()[0]["key"]
    assert j["meta"]["metrics"][0]["value"] == user_object.metrics()[0]["value"]


def test_model_bin_data():
    user_object = UserObject()
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()
    bdata = b"123"
    bdata_base64 = base64.b64encode(bdata).decode("utf-8")
    rv = client.get('/predict?json={"binData":"' + bdata_base64 + '"}')
    j = json.loads(rv.data)
    assert rv.status_code == 200
    assert j["binData"] == bdata_base64
    assert j["meta"]["tags"] == {"mytag": 1}
    assert j["meta"]["metrics"][0]["key"] == user_object.metrics()[0]["key"]
    assert j["meta"]["metrics"][0]["value"] == user_object.metrics()[0]["value"]


def test_model_bin_data_nparray():
    user_object = UserObject(ret_nparray=True)
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()
    encoded = base64.b64encode(b"1234").decode("utf-8")
    rv = client.get('/predict?json={"binData":"' + encoded + '"}')
    j = json.loads(rv.data)
    logging.info(j)
    assert rv.status_code == 200
    assert j["data"]["tensor"]["values"] == [1, 2, 3]
    assert j["meta"]["tags"] == {"mytag": 1}
    assert j["meta"]["metrics"][0]["key"] == user_object.metrics()[0]["key"]
    assert j["meta"]["metrics"][0]["value"] == user_object.metrics()[0]["value"]


def test_model_str_data():
    user_object = UserObject(ret_nparray=True)
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()
    rv = client.get('/predict?json={"strData":"my data"}')
    j = json.loads(rv.data)
    logging.info(j)
    assert rv.status_code == 200
    assert j["data"]["tensor"]["values"] == [1, 2, 3]
    assert j["meta"]["tags"] == {"mytag": 1}
    assert j["meta"]["metrics"][0]["key"] == user_object.metrics()[0]["key"]
    assert j["meta"]["metrics"][0]["value"] == user_object.metrics()[0]["value"]


def test_model_str_data_identity():
    user_object = UserObject()
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()
    rv = client.get('/predict?json={"strData":"my data"}')
    j = json.loads(rv.data)
    logging.info(j)
    assert rv.status_code == 200
    assert j["strData"] == "my data"
    assert j["meta"]["tags"] == {"mytag": 1}
    assert j["meta"]["metrics"][0]["key"] == user_object.metrics()[0]["key"]
    assert j["meta"]["metrics"][0]["value"] == user_object.metrics()[0]["value"]


def test_model_no_json():
    user_object = UserObject()
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()
    uo = UserObject()
    rv = client.get("/predict?", headers={"Content-Type": "application/json"})
    j = json.loads(rv.data)
    logging.info(j)
    assert rv.status_code == 400


def test_model_bad_metrics():
    user_object = UserObject(metrics_ok=False)
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()
    rv = client.get('/predict?json={"data":{"ndarray":[]}}')
    j = json.loads(rv.data)
    logging.info(j)
    assert rv.status_code == 500


def test_model_error_status_code():
    class ErrorUserObject:
        def predict(self, X, features_names, **kwargs):
            raise SeldonMicroserviceException("foo", status_code=403)

    user_object = ErrorUserObject()
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()
    uo = UserObject()
    rv = client.get('/predict?json={"strData":"my data"}')
    j = json.loads(rv.data)
    logging.info(j)
    assert rv.status_code == 403


def test_model_gets_meta():
    user_object = UserObject(ret_meta=True)
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()
    rv = client.get('/predict?json={"meta":{"puid": "abc"},"data":{"ndarray":[]}}')
    j = json.loads(rv.data)
    logging.info(j)

    assert rv.status_code == 200
    assert j["meta"]["tags"] == {"inc_meta": {"puid": "abc"}}
    assert j["meta"]["metrics"][0]["key"] == user_object.metrics()[0]["key"]
    assert j["meta"]["metrics"][0]["value"] == user_object.metrics()[0]["value"]


def test_model_passes_through_tags():
    user_object = UserObject()
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()
    rv = client.get(
        '/predict?json={"meta":{"tags":{"foo":"bar"}},"data":{"ndarray":[]}}'
    )
    j = json.loads(rv.data)
    logging.info(j)

    assert rv.status_code == 200
    assert j["meta"]["tags"] == {"foo": "bar", "mytag": 1}
    assert j["meta"]["metrics"][0]["key"] == user_object.metrics()[0]["key"]
    assert j["meta"]["metrics"][0]["value"] == user_object.metrics()[0]["value"]


def test_model_passes_through_metrics():
    user_object = UserObject()
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()
    rv = client.get(
        '/predict?json={"meta":{"metrics":[{"key":"request_gauge","type":"GAUGE","value":100}]},"data":{"ndarray":[]}}'
    )
    j = json.loads(rv.data)
    logging.info(j)

    assert rv.status_code == 200
    assert j["meta"]["metrics"][0]["key"] == "request_gauge"
    assert j["meta"]["metrics"][0]["value"] == 100

    assert j["meta"]["metrics"][1]["key"] == user_object.metrics()[0]["key"]
    assert j["meta"]["metrics"][1]["value"] == user_object.metrics()[0]["value"]


def test_model_seldon_json_ok():
    user_object = UserObject()
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()
    rv = client.get("/seldon.json")
    assert rv.status_code == 200


def test_model_health_ping():
    user_object = UserObject()
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()
    rv = client.get(HEALTH_PING_URL)
    assert rv.status_code == 200
    assert rv.data == b"pong"


def test_model_health_status():
    user_object = UserObject()
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()
    rv = client.get(HEALTH_STATUS_URL)
    assert rv.status_code == 200
    j = json.loads(rv.data)
    logging.info(j)
    assert j["data"]["tensor"]["values"] == UserObject.HEALTH_STATUS_RESPONSE


def test_model_health_status_raw():
    user_object = UserObjectLowLevel()
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()
    rv = client.get(HEALTH_STATUS_URL)
    assert rv.status_code == 200
    j = json.loads(rv.data)
    assert j["data"]["ndarray"] == UserObjectLowLevel.HEALTH_STATUS_RAW_RESPONSE


def test_model_metadata():
    user_object = UserObject()
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()
    rv = client.get(METADATA_URL)
    assert rv.status_code == 200
    j = json.loads(rv.data)
    logging.info(j)
    assert j == UserObject.METADATA_RESPONSE


def test_proto_ok():
    user_object = UserObject()
    seldon_metrics = SeldonMetrics()
    app = SeldonModelGRPC(user_object, seldon_metrics)
    arr = np.array([1, 2])
    datadef = prediction_pb2.DefaultData(
        tensor=prediction_pb2.Tensor(shape=(2, 1), values=arr)
    )
    request = prediction_pb2.SeldonMessage(data=datadef)
    resp = app.Predict(request, None)
    jStr = json_format.MessageToJson(resp)
    j = json.loads(jStr)
    logging.info(j)
    assert j["meta"]["tags"] == {"mytag": 1}
    assert j["meta"]["metrics"][0]["key"] == user_object.metrics()[0]["key"]
    assert j["meta"]["metrics"][0]["value"] == user_object.metrics()[0]["value"]
    assert j["data"]["tensor"]["shape"] == [2, 1]
    assert j["data"]["tensor"]["values"] == [1, 2]


def test_proto_passes_through_tags():
    user_object = UserObject()
    seldon_metrics = SeldonMetrics()
    app = SeldonModelGRPC(user_object, seldon_metrics)
    arr = np.array([1, 2])
    datadef = prediction_pb2.DefaultData(
        tensor=prediction_pb2.Tensor(shape=(2, 1), values=arr)
    )
    meta = prediction_pb2.Meta()
    json_format.ParseDict({"tags": {"foo": "bar"}}, meta)
    request = prediction_pb2.SeldonMessage(data=datadef, meta=meta)
    resp = app.Predict(request, None)
    jStr = json_format.MessageToJson(resp)
    j = json.loads(jStr)
    logging.info(j)
    assert j["meta"]["tags"] == {"foo": "bar", "mytag": 1}
    assert j["meta"]["metrics"][0]["key"] == user_object.metrics()[0]["key"]
    assert j["meta"]["metrics"][0]["value"] == user_object.metrics()[0]["value"]
    assert j["data"]["tensor"]["shape"] == [2, 1]
    assert j["data"]["tensor"]["values"] == [1, 2]


def test_proto_passes_through_metrics():
    user_object = UserObject()
    seldon_metrics = SeldonMetrics()
    app = SeldonModelGRPC(user_object, seldon_metrics)
    arr = np.array([1, 2])
    datadef = prediction_pb2.DefaultData(
        tensor=prediction_pb2.Tensor(shape=(2, 1), values=arr)
    )
    meta = prediction_pb2.Meta()
    json_format.ParseDict(
        {"metrics": [{"key": "request_gauge", "type": "GAUGE", "value": 100}]}, meta
    )
    request = prediction_pb2.SeldonMessage(data=datadef, meta=meta)
    resp = app.Predict(request, None)
    jStr = json_format.MessageToJson(resp)
    j = json.loads(jStr)
    logging.info(j)

    assert j["meta"]["metrics"][0]["key"] == "request_gauge"
    assert j["meta"]["metrics"][0]["value"] == 100

    assert j["meta"]["metrics"][1]["key"] == user_object.metrics()[0]["key"]
    assert j["meta"]["metrics"][1]["value"] == user_object.metrics()[0]["value"]

    assert j["data"]["tensor"]["shape"] == [2, 1]
    assert j["data"]["tensor"]["values"] == [1, 2]


def test_proto_lowlevel():
    user_object = UserObjectLowLevelGrpc()
    seldon_metrics = SeldonMetrics()
    app = SeldonModelGRPC(user_object, seldon_metrics)
    arr = np.array([1, 2])
    datadef = prediction_pb2.DefaultData(
        tensor=prediction_pb2.Tensor(shape=(2, 1), values=arr)
    )
    request = prediction_pb2.SeldonMessage(data=datadef)
    resp = app.Predict(request, None)
    jStr = json_format.MessageToJson(resp)
    j = json.loads(jStr)
    logging.info(j)
    assert j["data"]["tensor"]["shape"] == [2, 1]
    assert j["data"]["tensor"]["values"] == [9, 9]


def test_proto_feedback():
    user_object = UserObject()
    seldon_metrics = SeldonMetrics()
    app = SeldonModelGRPC(user_object, seldon_metrics)
    arr = np.array([1, 2])
    datadef = prediction_pb2.DefaultData(
        tensor=prediction_pb2.Tensor(shape=(2, 1), values=arr)
    )
    request = prediction_pb2.SeldonMessage(data=datadef)
    feedback = prediction_pb2.Feedback(request=request, reward=1.0)
    resp = app.SendFeedback(feedback, None)


def test_proto_feedback_custom():
    user_object = UserObjectLowLevel()
    seldon_metrics = SeldonMetrics()
    app = SeldonModelGRPC(user_object, seldon_metrics)
    arr = np.array([1, 2])
    datadef = prediction_pb2.DefaultData(
        tensor=prediction_pb2.Tensor(shape=(2, 1), values=arr)
    )
    request = prediction_pb2.SeldonMessage(data=datadef)
    feedback = prediction_pb2.Feedback(request=request, reward=1.0)
    resp = app.SendFeedback(feedback, None)


@skipif_tf_missing
def test_proto_tftensor_ok():
    user_object = UserObject()
    seldon_metrics = SeldonMetrics()
    app = SeldonModelGRPC(user_object, seldon_metrics)
    arr = np.array([1, 2])
    datadef = prediction_pb2.DefaultData(tftensor=tf.make_tensor_proto(arr))
    request = prediction_pb2.SeldonMessage(data=datadef)
    resp = app.Predict(request, None)
    jStr = json_format.MessageToJson(resp)
    j = json.loads(jStr)
    logging.info(j)
    assert j["meta"]["tags"] == {"mytag": 1}
    assert j["meta"]["metrics"][0]["key"] == user_object.metrics()[0]["key"]
    assert j["meta"]["metrics"][0]["value"] == user_object.metrics()[0]["value"]
    arr2 = tf.make_ndarray(resp.data.tftensor)
    assert np.array_equal(arr, arr2)


def test_proto_bin_data():
    user_object = UserObject()
    seldon_metrics = SeldonMetrics()
    app = SeldonModelGRPC(user_object, seldon_metrics)
    bdata = b"123"
    bdata_base64 = base64.b64encode(bdata)
    request = prediction_pb2.SeldonMessage(binData=bdata_base64)
    resp = app.Predict(request, None)
    assert resp.binData == bdata_base64


def test_proto_bin_data_nparray():
    user_object = UserObject(ret_nparray=True)
    seldon_metrics = SeldonMetrics()
    app = SeldonModelGRPC(user_object, seldon_metrics)
    binData = b"\0\1"
    request = prediction_pb2.SeldonMessage(binData=binData)
    resp = app.Predict(request, None)
    jStr = json_format.MessageToJson(resp)
    j = json.loads(jStr)
    logging.info(j)
    assert j["data"]["tensor"]["values"] == list(user_object.nparray.flatten())


def test_get_grpc_server():
    user_object = UserObject(ret_nparray=True)
    seldon_metrics = SeldonMetrics()
    server = get_grpc_server(user_object, seldon_metrics)


def test_proto_gets_meta():
    user_object = UserObject(ret_meta=True)
    seldon_metrics = SeldonMetrics()
    app = SeldonModelGRPC(user_object, seldon_metrics)
    arr = np.array([1, 2])
    datadef = prediction_pb2.DefaultData(
        tensor=prediction_pb2.Tensor(shape=(2, 1), values=arr)
    )
    meta = prediction_pb2.Meta()
    metaJson = {"puid": "abc"}
    json_format.ParseDict(metaJson, meta)
    request = prediction_pb2.SeldonMessage(data=datadef, meta=meta)
    resp = app.Predict(request, None)
    jStr = json_format.MessageToJson(resp)
    j = json.loads(jStr)
    logging.info(j)
    assert j["meta"]["tags"] == {"inc_meta": {"puid": "abc"}}
    assert j["meta"]["metrics"][0]["key"] == user_object.metrics()[0]["key"]
    assert j["meta"]["metrics"][0]["value"] == user_object.metrics()[0]["value"]
    assert j["data"]["tensor"]["shape"] == [2, 1]
    assert j["data"]["tensor"]["values"] == [1, 2]


def test_unimplemented_predict_raw_on_seldon_component():
    class CustomSeldonComponent(SeldonComponent):
        def predict(self, X, features_names, **kwargs):
            return X * 2

    user_object = CustomSeldonComponent()
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()
    rv = client.get('/predict?json={"data":{"names":["a","b"],"ndarray":[[1,2]]}}')
    j = json.loads(rv.data)

    logging.info(j)
    assert rv.status_code == 200
    assert j["data"]["ndarray"] == [[2.0, 4.0]]


def test_unimplemented_predict_raw():
    class CustomObject:
        def predict(self, X, features_names, **kwargs):
            return X * 2

    user_object = CustomObject()
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()
    rv = client.get('/predict?json={"data":{"names":["a","b"],"ndarray":[[1,2]]}}')
    j = json.loads(rv.data)

    logging.info(j)
    assert rv.status_code == 200
    assert j["data"]["ndarray"] == [[2.0, 4.0]]


def test_unimplemented_feedback_raw_on_seldon_component():
    class CustomSeldonComponent(SeldonComponent):
        def feedback(self, features, feature_names, reward, truth):
            logging.info("Feedback called")

    user_object = CustomSeldonComponent()
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()
    rv = client.get(
        '/send-feedback?json={"request":{"data":{"ndarray":[]}},"reward":1.0}'
    )
    j = json.loads(rv.data)

    logging.info(j)
    assert rv.status_code == 200


def test_unimplemented_feedback_raw():
    class CustomObject:
        def feedback(self, features, feature_names, reward, truth):
            logging.info("Feedback called")

    user_object = CustomObject()
    seldon_metrics = SeldonMetrics()
    app = get_rest_microservice(user_object, seldon_metrics)
    client = app.test_client()
    rv = client.get(
        '/send-feedback?json={"request":{"data":{"ndarray":[]}},"reward":1.0}'
    )
    j = json.loads(rv.data)

    logging.info(j)
    assert rv.status_code == 200



import http.client as http_client
import json
import logging
from typing import Dict, Iterable, List, Optional, Tuple, Union

import grpc
import numpy as np
import requests
from google.protobuf import any_pb2, json_format

from seldon_core.proto import prediction_pb2, prediction_pb2_grpc
from seldon_core.utils import (
    array_to_grpc_datadef,
    feedback_to_json,
    json_to_feedback,
    json_to_seldon_message,
    seldon_message_to_json,
    seldon_messages_to_json,
)

logger = logging.getLogger(__name__)


class SeldonClientException(Exception):
    """
    Seldon Client Exception
    """

    status_code = 400

    def __init__(self, message):
        Exception.__init__(self)
        self.message = message


class SeldonChannelCredentials:
    """
    Channel credentials.

    Presently just denotes an SSL connection.
    For GRPC in order to be properly implemented, you need to provide *either*
    the root_certificate_files, *or* all the file paths.
    The verify attribute currently is used to avoid SSL verification in REST
    however for GRPC it is recommended that you provide a path at least for the
    root_certificates_file otherwise it may not work as expected.
    """

    def __init__(
        self,
        verify: bool = True,
        root_certificates_file: str = None,
        private_key_file: str = None,
        certificate_chain_file: str = None,
    ):
        self.verify = verify
        self.root_certificates_file = root_certificates_file
        self.private_key_file = private_key_file
        self.certificate_chain_file = certificate_chain_file


class SeldonCallCredentials:
    """
    Credentials for each call, currently implements the ability to provide
        an OAuth token which is currently made available through REST via
        the X-Auth-Token header, and via GRPC via the metadata call creds.
    """

    def __init__(self, token: str = None):
        self.token = token


class SeldonClientPrediction:
    """
    Data class to return from Seldon Client
    """

    def __init__(
        self,
        request: Optional[Union[prediction_pb2.SeldonMessage, Dict]],
        response: Optional[Union[prediction_pb2.SeldonMessage, Dict]],
        success: bool = True,
        msg: str = "",
    ):
        self.request = request
        self.response = response
        self.success = success
        self.msg = msg

    def __repr__(self):
        return "Success:%s message:%s\nRequest:\n%s\nResponse:\n%s" % (
            self.success,
            self.msg,
            self.request,
            self.response,
        )


class SeldonClientFeedback:
    """
    Data class to return from Seldon Client for feedback calls
    """

    def __init__(
        self,
        request: Optional[prediction_pb2.Feedback],
        response: Optional[Union[prediction_pb2.SeldonMessage, Dict]],
        success: bool = True,
        msg: str = "",
    ):
        self.request = request
        self.response = response
        self.success = success
        self.msg = msg

    def __repr__(self):
        return "Success:%s message:%s\nRequest:\n%s\nResponse:\n%s" % (
            self.success,
            self.msg,
            self.request,
            self.response,
        )


class SeldonClientCombine:
    """
    Data class to return from Seldon Client for aggregate calls
    """

    def __init__(
        self,
        request: Optional[prediction_pb2.SeldonMessageList],
        response: Optional[prediction_pb2.SeldonMessage],
        success: bool = True,
        msg: str = "",
    ):
        self.request = request
        self.response = response
        self.success = success
        self.msg = msg

    def __repr__(self):
        return "Success:%s message:%s\nRequest:\n%s\nResponse:\n%s" % (
            self.success,
            self.msg,
            self.request,
            self.response,
        )


class SeldonClient:
    """
    A reference Seldon API Client
    """

    def __init__(
        self,
        gateway: str = "ambassador",
        transport: str = "rest",
        namespace: str = None,
        deployment_name: str = None,
        payload_type: str = "tensor",
        gateway_endpoint: str = "localhost:8003",
        microservice_endpoint: str = "localhost:5000",
        grpc_max_send_message_length: int = 4 * 1024 * 1024,
        grpc_max_receive_message_length: int = 4 * 1024 * 1024,
        channel_credentials: SeldonChannelCredentials = None,
        call_credentials: SeldonCallCredentials = None,
        debug: bool = False,
        client_return_type: str = "dict",
        ssl: bool = None,
    ):
        """

        Parameters
        ----------
        gateway
           API Gateway - either ambassador, istio or seldon
        transport
           API transport - grpc or rest
        namespace
           k8s namespace of running deployment
        deployment_name
           name of seldon deployment
        payload_type
           type of payload - tensor, ndarray or tftensor
        gateway_endpoint
           Gateway endpoint
        microservice_endpoint
           Running microservice endpoint
        grpc_max_send_message_length
           Max grpc send message size in bytes
        grpc_max_receive_message_length
           Max grpc receive message size in bytes
        client_return_type
            the return type of all functions can be either dict or proto
        """
        if debug:
            logger.setLevel(logging.DEBUG)
            http_client.HTTPConnection.debuglevel = 1
        self.config = locals().copy()
        del self.config["self"]
        logger.debug("Configuration:" + str(self.config))

    def _gather_args(self, **kwargs):
        c2 = {**self.config}
        c2.update({k: v for k, v in kwargs.items() if v is not None})
        return c2

    def _validate_args(
        self,
        gateway: str = None,
        transport: str = None,
        method: str = None,
        data: np.ndarray = None,
        client_return_type: str = "dict",
        **kwargs,
    ):
        """
        Internal method to validate parameters

        Parameters
        ----------
        gateway
           API gateway
        transport
           API transport
        method
           The method to call
        data
           Numpy data to send
        kwargs

        Returns
        -------

        """
        if not (gateway == "ambassador" or gateway == "seldon" or gateway == "istio"):
            raise SeldonClientException(
                "Valid values for gateway are 'ambassador', 'istio', or 'seldon'"
            )
        if not (transport == "rest" or transport == "grpc"):
            raise SeldonClientException(
                "Valid values for transport are 'rest' or 'grpc'"
            )
        if not (
            method == "predict"
            or method == "route"
            or method == "aggregate"
            or method == "transform-input"
            or method == "transform-output"
            or method == "send-feedback"
            or method is None
        ):
            raise SeldonClientException(
                "Valid values for method are 'predict', 'route', 'transform-input', 'transform-output', 'aggregate' or None"
            )
        if not (data is None or isinstance(data, np.ndarray)):
            raise SeldonClientException("Valid values for data are None or numpy array")
        if not (client_return_type == "proto" or client_return_type == "dict"):
            raise SeldonClientException(
                "Valid values for client_return_type are proto or dict"
            )

    def predict(
        self,
        gateway: str = None,
        transport: str = None,
        deployment_name: str = None,
        payload_type: str = None,
        gateway_endpoint: str = None,
        microservice_endpoint: str = None,
        method: str = None,
        shape: Tuple = (1, 1),
        namespace: str = None,
        data: np.ndarray = None,
        bin_data: Union[bytes, bytearray] = None,
        str_data: str = None,
        json_data: Union[str, List, Dict] = None,
        custom_data: any_pb2.Any = None,
        names: Iterable[str] = None,
        gateway_prefix: str = None,
        headers: Dict = None,
        http_path: str = None,
        meta: Dict = None,
        client_return_type: str = None,
        raw_data: Dict = None,
        ssl: bool = None,
    ) -> SeldonClientPrediction:
        """

        Parameters
        ----------
        gateway
           API Gateway - either ambassador, istio or seldon
        transport
           API transport - grpc or rest
        namespace
           k8s namespace of running deployment
        deployment_name
           name of seldon deployment
        payload_type
           type of payload - tensor, ndarray or tftensor
        gateway_endpoint
           Gateway endpoint
        microservice_endpoint
           Running microservice endpoint
        grpc_max_send_message_length
           Max grpc send message size in bytes
        grpc_max_receive_message_length
           Max grpc receive message size in bytes
        data
           Numpy Array Payload to send
        bin_data
           Binary payload to send - will override data
        str_data
           String payload to send - will override data
        json_data
           JSON payload to send - will override data
        custom_data
           Custom payload to send - will override data
        names
           Column names
        gateway_prefix
           prefix path for gateway URL endpoint
        headers
           Headers to add to request
        http_path:
           Custom http path for predict call to use
        meta:
           Custom meta map, supplied as tags
        client_return_type
            the return type of all functions can be either dict or proto
        raw_data
            Raw payload, a dictionary representing the json request or the raw grpc proto

        Returns
        -------

        """
        k = self._gather_args(
            gateway=gateway,
            transport=transport,
            deployment_name=deployment_name,
            payload_type=payload_type,
            gateway_endpoint=gateway_endpoint,
            microservice_endpoint=microservice_endpoint,
            method=method,
            shape=shape,
            namespace=namespace,
            names=names,
            data=data,
            bin_data=bin_data,
            str_data=str_data,
            json_data=json_data,
            custom_data=custom_data,
            gateway_prefix=gateway_prefix,
            headers=headers,
            http_path=http_path,
            meta=meta,
            client_return_type=client_return_type,
            raw_data=raw_data,
            ssl=ssl,
        )
        self._validate_args(**k)
        if k["gateway"] == "ambassador" or k["gateway"] == "istio":
            if k["transport"] == "rest":
                return rest_predict_gateway(**k)
            elif k["transport"] == "grpc":
                return grpc_predict_gateway(**k)
            else:
                raise SeldonClientException("Unknown transport " + k["transport"])
        elif k["gateway"] == "seldon":
            if k["transport"] == "rest":
                return rest_predict_seldon(**k)
            elif k["transport"] == "grpc":
                return grpc_predict_seldon(**k)
            else:
                raise SeldonClientException("Unknown transport " + k["transport"])
        else:
            raise SeldonClientException("Unknown gateway " + k["gateway"])

    def feedback(
        self,
        prediction_request: prediction_pb2.SeldonMessage = None,
        prediction_response: prediction_pb2.SeldonMessage = None,
        prediction_truth: prediction_pb2.SeldonMessage = None,
        reward: float = 0,
        gateway: str = None,
        transport: str = None,
        deployment_name: str = None,
        payload_type: str = None,
        gateway_endpoint: str = None,
        microservice_endpoint: str = None,
        method: str = None,
        shape: Tuple = (1, 1),
        namespace: str = None,
        gateway_prefix: str = None,
        client_return_type: str = None,
        raw_request: dict = None,
        ssl: bool = None,
    ) -> SeldonClientFeedback:
        """

        Parameters
        ----------
        prediction_request
           Previous prediction request
        prediction_response
           Previous prediction response
        reward
           A reward to send in feedback
        gateway
           API Gateway - either ambassador, istio or seldon
        transport
           API transport - grpc or rest
        deployment_name
           name of seldon deployment
        payload_type
           payload - tensor, ndarray or tftensor
        gateway_endpoint
           Gateway endpoint
        microservice_endpoint
           Running microservice endpoint
        grpc_max_send_message_length
           Max grpc send message size in bytes
        grpc_max_receive_message_length
           Max grpc receive message size in bytes
        method
           The microservice method to call
        shape
           The shape of the data to send
        namespace
           k8s namespace of running deployment
        client_return_type
            the return type of all functions can be either dict or proto

        Returns
        -------

        """
        k = self._gather_args(
            gateway=gateway,
            transport=transport,
            deployment_name=deployment_name,
            payload_type=payload_type,
            gateway_endpoint=gateway_endpoint,
            microservice_endpoint=microservice_endpoint,
            method=method,
            shape=shape,
            namespace=namespace,
            gateway_prefix=gateway_prefix,
            client_return_type=client_return_type,
            raw_request=raw_request,
            ssl=ssl,
        )
        self._validate_args(**k)
        if k["gateway"] == "ambassador" or k["gateway"] == "istio":
            if k["transport"] == "rest":
                return rest_feedback_gateway(
                    prediction_request,
                    prediction_response,
                    prediction_truth,
                    reward,
                    **k,
                )
            elif k["transport"] == "grpc":
                return grpc_feedback_gateway(
                    prediction_request,
                    prediction_response,
                    prediction_truth,
                    reward,
                    **k,
                )
            else:
                raise SeldonClientException("Unknown transport " + k["transport"])
        elif k["gateway"] == "seldon":
            if k["transport"] == "rest":
                return rest_feedback_seldon(
                    prediction_request,
                    prediction_response,
                    prediction_truth,
                    reward,
                    **k,
                )
            elif k["transport"] == "grpc":
                return grpc_feedback_seldon(
                    prediction_request,
                    prediction_response,
                    prediction_truth,
                    reward,
                    **k,
                )
            else:
                raise SeldonClientException("Unknown transport " + k["transport"])
        else:
            raise SeldonClientException("Unknown gateway " + k["gateway"])

    def explain(
        self,
        gateway: str = None,
        transport: str = None,
        deployment_name: str = None,
        payload_type: str = None,
        gateway_endpoint: str = None,
        shape: Tuple = (1, 1),
        namespace: str = None,
        data: np.ndarray = None,
        bin_data: Union[bytes, bytearray] = None,
        str_data: str = None,
        json_data: str = None,
        names: Iterable[str] = None,
        gateway_prefix: str = None,
        headers: Dict = None,
        http_path: str = None,
        client_return_type: str = None,
        predictor: str = None,
        ssl: bool = None,
    ) -> Dict:
        """

        Parameters
        ----------
        gateway
           API Gateway - either ambassador, istio or seldon
        transport
           API transport - grpc or rest
        namespace
           k8s namespace of running deployment
        deployment_name
           name of seldon deployment
        payload_type
           type of payload - tensor, ndarray or tftensor
        gateway_endpoint
           Gateway endpoint
        microservice_endpoint
           Running microservice endpoint
        grpc_max_send_message_length
           Max grpc send message size in bytes
        grpc_max_receive_message_length
           Max grpc receive message size in bytes
        data
           Numpy Array Payload to send
        bin_data
           Binary payload to send - will override data
        str_data
           String payload to send - will override data
        json_data
           JSON payload to send - will override data
        names
           Column names
        gateway_prefix
           prefix path for gateway URL endpoint
        headers
           Headers to add to request
        http_path:
           Custom http path for predict call to use
        client_return_type
            the return type of all functions can be either dict or proto
        predictor
            The name of the predictor to send the explanations to

        Returns
        -------

        """
        k = self._gather_args(
            gateway=gateway,
            transport=transport,
            deployment_name=deployment_name,
            payload_type=payload_type,
            gateway_endpoint=gateway_endpoint,
            shape=shape,
            namespace=namespace,
            names=names,
            data=data,
            bin_data=bin_data,
            str_data=str_data,
            json_data=json_data,
            gateway_prefix=gateway_prefix,
            headers=headers,
            http_path=http_path,
            client_return_type=client_return_type,
            predictor=predictor,
            ssl=ssl,
        )
        self._validate_args(**k)
        if k["gateway"] == "ambassador" or k["gateway"] == "istio":
            if k["transport"] == "rest":
                return explain_predict_gateway(**k)
            elif k["transport"] == "grpc":
                raise SeldonClientException("gRPC not supported for explain")
            else:
                raise SeldonClientException("Unknown transport " + k["transport"])
        else:
            raise SeldonClientException("Unknown gateway " + k["gateway"])

    def microservice(
        self,
        gateway: str = None,
        transport: str = None,
        deployment_name: str = None,
        payload_type: str = None,
        gateway_endpoint: str = None,
        microservice_endpoint: str = None,
        method: str = None,
        shape: Tuple = (1, 1),
        namespace: str = None,
        data: np.ndarray = None,
        datas: List[np.ndarray] = None,
        ndatas: int = None,
        bin_data: Union[bytes, bytearray] = None,
        str_data: str = None,
        json_data: Union[str, List, Dict] = None,
        custom_data: any_pb2.Any = None,
        names: Iterable[str] = None,
    ) -> Union[SeldonClientPrediction, SeldonClientCombine]:
        """

        Parameters
        ----------
        gateway
           API Gateway - either ambassador, istio or seldon
        transport
           API transport - grpc or rest
        deployment_name
           name of seldon deployment
        payload_type
           payload - tensor, ndarray or tftensor
        gateway_endpoint
           Gateway endpoint
        microservice_endpoint
           Running microservice endpoint
        grpc_max_send_message_length
           Max grpc send message size in bytes
        grpc_max_receive_message_length
           Max grpc receive message size in bytes
        method
           The microservice method to call
        shape
           The shape of the data to send
        namespace
           k8s namespace of running deployment
        data
           Numpy Array Payload to send
        bin_data
           Binary payload to send - will override data
        str_data
           String payload to send - will override data
        json_data
           String payload to send - will override data
        custom_data
           Custom payload to send - will override data
        ndatas
           Multiple numpy arrays to send for aggregation
        bin_data
           Binary data payload
        str_data
           String data payload
        names
           Column names

        Returns
        -------
           A prediction result

        """
        k = self._gather_args(
            gateway=gateway,
            transport=transport,
            deployment_name=deployment_name,
            payload_type=payload_type,
            gateway_endpoint=gateway_endpoint,
            microservice_endpoint=microservice_endpoint,
            method=method,
            shape=shape,
            namespace=namespace,
            datas=datas,
            ndatas=ndatas,
            names=names,
            data=data,
            bin_data=bin_data,
            str_data=str_data,
            json_data=json_data,
            custom_data=custom_data,
        )
        self._validate_args(**k)
        if k["transport"] == "rest":
            if (
                k["method"] == "predict"
                or k["method"] == "transform-input"
                or k["method"] == "transform-output"
                or k["method"] == "route"
            ):
                return microservice_api_rest_seldon_message(**k)
            elif k["method"] == "aggregate":
                return microservice_api_rest_aggregate(**k)
            else:
                raise SeldonClientException("Unknown method " + k["method"])
        elif k["transport"] == "grpc":
            if (
                k["method"] == "predict"
                or k["method"] == "transform-input"
                or k["method"] == "transform-output"
                or k["method"] == "route"
            ):
                return microservice_api_grpc_seldon_message(**k)
            elif k["method"] == "aggregate":
                return microservice_api_grpc_aggregate(**k)
            else:
                raise SeldonClientException("Unknown method " + k["method"])
        else:
            raise SeldonClientException("Unknown transport " + k["transport"])

    def microservice_feedback(
        self,
        prediction_request: prediction_pb2.SeldonMessage = None,
        prediction_response: prediction_pb2.SeldonMessage = None,
        reward: float = 0,
        gateway: str = None,
        transport: str = None,
        deployment_name: str = None,
        payload_type: str = None,
        gateway_endpoint: str = None,
        microservice_endpoint: str = None,
        method: str = None,
        shape: Tuple = (1, 1),
        namespace: str = None,
    ) -> SeldonClientFeedback:
        """

        Parameters
        ----------
        prediction_request
           Previous prediction request
        prediction_response
           Previous prediction response
        reward
           A reward to send in feedback
        gateway
           API Gateway - either Gateway or seldon
        transport
           API transport - grpc or rest
        deployment_name
           name of seldon deployment
        payload_type
           payload - tensor, ndarray or tftensor
        gateway_endpoint
           Gateway endpoint
        microservice_endpoint
           Running microservice endpoint
        grpc_max_send_message_length
           Max grpc send message size in bytes
        grpc_max_receive_message_length
           Max grpc receive message size in bytes
        method
           The microservice method to call
        shape
           The shape of the data to send
        namespace
           k8s namespace of running deployment

        Returns
        -------
           A client response

        """
        k = self._gather_args(
            gateway=gateway,
            transport=transport,
            deployment_name=deployment_name,
            payload_type=payload_type,
            gateway_endpoint=gateway_endpoint,
            microservice_endpoint=microservice_endpoint,
            method=method,
            shape=shape,
            namespace=namespace,
        )
        self._validate_args(**k)
        if k["transport"] == "rest":
            return microservice_api_rest_feedback(
                prediction_request, prediction_response, reward, **k
            )
        else:
            return microservice_api_grpc_feedback(
                prediction_request, prediction_response, reward, **k
            )


def microservice_api_rest_seldon_message(
    method: str = "predict",
    microservice_endpoint: str = "localhost:5000",
    shape: Tuple = (1, 1),
    data: object = None,
    payload_type: str = "tensor",
    bin_data: Union[bytes, bytearray] = None,
    str_data: str = None,
    json_data: Union[str, List, Dict] = None,
    names: Iterable[str] = None,
    **kwargs,
) -> SeldonClientPrediction:
    """
    Call Seldon microservice REST API

    Parameters
    ----------
    method
       The microservice method to call
    microservice_endpoint
       Running microservice endpoint
    grpc_max_send_message_length
       Max grpc send message size in bytes
    grpc_max_receive_message_length
       Max grpc receive message size in bytes
    method
       The microservice method to call
    shape
       The shape of the data to send
    namespace
       k8s namespace of running deployment
    shape
       Shape of the data to send
    data
       Numpy array data to send
    payload_type
       payload - tensor, ndarray or tftensor
    bin_data
       Binary data payload
    str_data
       String data payload
    json_data
       JSON data payload
    names
       Column names
    kwargs

    Returns
    -------
      A SeldonClientPrediction data response
    """
    if bin_data is not None:
        request = prediction_pb2.SeldonMessage(binData=bin_data)
    elif str_data is not None:
        request = prediction_pb2.SeldonMessage(strData=str_data)
    elif json_data is not None:
        request = json_to_seldon_message({"jsonData": json_data})
    else:
        if data is None:
            data = np.random.rand(*shape)
        datadef = array_to_grpc_datadef(payload_type, data, names=names)
        request = prediction_pb2.SeldonMessage(data=datadef)
    payload = seldon_message_to_json(request)
    response_raw = requests.post(
        "http://" + microservice_endpoint + "/" + method,
        data={"json": json.dumps(payload)},
    )
    if response_raw.status_code == 200:
        success = True
        msg = ""
    else:
        success = False
        msg = response_raw.reason
    try:
        response = json_to_seldon_message(response_raw.json())
        return SeldonClientPrediction(request, response, success, msg)
    except Exception as e:
        return SeldonClientPrediction(request, None, success, str(e))


def microservice_api_rest_aggregate(
    microservice_endpoint: str = "localhost:5000",
    shape: Tuple = (1, 1),
    datas: List[np.ndarray] = None,
    ndatas: int = None,
    payload_type: str = "tensor",
    names: Iterable[str] = None,
    **kwargs,
) -> SeldonClientCombine:
    """
    Call Seldon microservice REST API aggregate endpoint

    Parameters
    ----------
    microservice_endpoint
       Running microservice endpoint
    shape
       The shape of the data to send
    datas
       List of Numpy array data to send
    ndatas
       Multiple numpy arrays to send for aggregation
    payload_type
       payload - tensor, ndarray or tftensor
    names
       Column names
    kwargs

    Returns
    -------
       A SeldonClientPrediction

    """
    if datas is None:
        datas = []
        for n in range(ndatas):
            data = np.random.rand(*shape)
            datas.append(data)
    msgs = []
    for data in datas:
        if isinstance(data, (bytes, bytearray)):
            msgs.append(prediction_pb2.SeldonMessage(binData=data))
        elif isinstance(data, str):
            msgs.append(prediction_pb2.SeldonMessage(strData=data))
        else:
            datadef = array_to_grpc_datadef(payload_type, data, names)
            msgs.append(prediction_pb2.SeldonMessage(data=datadef))
    request = prediction_pb2.SeldonMessageList(seldonMessages=msgs)
    payload = seldon_messages_to_json(request)
    response_raw = requests.post(
        "http://" + microservice_endpoint + "/aggregate",
        data={"json": json.dumps(payload)},
    )
    if response_raw.status_code == 200:
        success = True
        msg = ""
    else:
        success = False
        msg = response_raw.reason
    try:
        response = json_to_seldon_message(response_raw.json())
        return SeldonClientCombine(request, response, success, msg)
    except Exception as e:
        return SeldonClientCombine(request, None, success, str(e))


def microservice_api_rest_feedback(
    prediction_request: prediction_pb2.SeldonMessage = None,
    prediction_response: prediction_pb2.SeldonMessage = None,
    reward: float = 0,
    microservice_endpoint: str = None,
    **kwargs,
) -> SeldonClientFeedback:
    """
    Call Seldon microserice REST API to send feedback

    Parameters
    ----------
    prediction_request
       Previous prediction request
    prediction_response
       Previous prediction response
    reward
       A reward to send in feedback
    microservice_endpoint
       Running microservice endpoint
    kwargs

    Returns
    -------
       A SeldonClientFeedback
    """
    request = prediction_pb2.Feedback(
        request=prediction_request, response=prediction_response, reward=reward
    )
    payload = feedback_to_json(request)
    response_raw = requests.post(
        "http://" + microservice_endpoint + "/send-feedback",
        data={"json": json.dumps(payload)},
    )
    if response_raw.status_code == 200:
        success = True
        msg = ""
    else:
        success = False
        msg = response_raw.reason
    try:
        response = json_to_seldon_message(response_raw.json())
        return SeldonClientFeedback(request, response, success, msg)
    except Exception as e:
        return SeldonClientFeedback(request, None, success, str(e))


def microservice_api_grpc_seldon_message(
    method: str = "predict",
    microservice_endpoint: str = "localhost:5000",
    shape: Tuple = (1, 1),
    data: object = None,
    payload_type: str = "tensor",
    bin_data: Union[bytes, bytearray] = None,
    str_data: str = None,
    json_data: Union[str, List, Dict] = None,
    custom_data: any_pb2.Any = None,
    grpc_max_send_message_length: int = 4 * 1024 * 1024,
    grpc_max_receive_message_length: int = 4 * 1024 * 1024,
    names: Iterable[str] = None,
    **kwargs,
) -> SeldonClientPrediction:
    """
    Call Seldon microservice gRPC API

    Parameters
    ----------
    method
       Method to call
    microservice_endpoint
       Running microservice endpoint
    shape
       The shape of the data to send
    data
       Numpy array data to send
    payload_type
       payload - tensor, ndarray or tftensor
    bin_data
       Binary data to send
    str_data
       String data to send
    json_data
        JSON data to send
    custom_data
        Custom data to send
    grpc_max_send_message_length
       Max grpc send message size in bytes
    grpc_max_receive_message_length
       Max grpc receive message size in bytes
    names
       column names
    kwargs

    Returns
    -------
      SeldonClientPrediction
    """
    if bin_data is not None:
        request = prediction_pb2.SeldonMessage(binData=bin_data)
    elif str_data is not None:
        request = prediction_pb2.SeldonMessage(strData=str_data)
    elif json_data is not None:
        request = json_to_seldon_message({"jsonData": json_data})
    elif custom_data is not None:
        request = prediction_pb2.SeldonMessage(customData=custom_data)
    else:
        if data is None:
            data = np.random.rand(*shape)
        datadef = array_to_grpc_datadef(payload_type, data, names=names)
        request = prediction_pb2.SeldonMessage(data=datadef)
    channel = grpc.insecure_channel(
        microservice_endpoint,
        options=[
            ("grpc.max_send_message_length", grpc_max_send_message_length),
            ("grpc.max_receive_message_length", grpc_max_receive_message_length),
        ],
    )
    try:
        if method == "predict":
            stub_model = prediction_pb2_grpc.ModelStub(channel)
            response = stub_model.Predict(request=request)
        elif method == "transform-input":
            stub = prediction_pb2_grpc.GenericStub(channel)
            response = stub.TransformInput(request=request)
        elif method == "transform-output":
            stub = prediction_pb2_grpc.GenericStub(channel)
            response = stub.TransformOutput(request=request)
        elif method == "route":
            stub = prediction_pb2_grpc.GenericStub(channel)
            response = stub.Route(request=request)
        else:
            raise SeldonClientException("Unknown method:" + method)

        channel.close()
        return SeldonClientPrediction(request, response, True, "")
    except Exception as e:
        channel.close()
        return SeldonClientPrediction(request, None, False, str(e))


def microservice_api_grpc_aggregate(
    microservice_endpoint: str = "localhost:5000",
    shape: Tuple = (1, 1),
    datas: List[np.ndarray] = None,
    ndatas: int = None,
    payload_type: str = "tensor",
    grpc_max_send_message_length: int = 4 * 1024 * 1024,
    grpc_max_receive_message_length: int = 4 * 1024 * 1024,
    names: Iterable[str] = None,
    **kwargs,
) -> SeldonClientCombine:
    """
    Call Seldon microservice gRPC API aggregate

    Parameters
    ----------
    microservice_endpoint
       Microservice API endpoint
    shape
       Shape of the data to send
    datas
       List of Numpy array data to send
    ndatas
       Multiple numpy arrays to send for aggregation
    payload_type
       payload - tensor, ndarray or tftensor
    grpc_max_send_message_length
       Max grpc send message size in bytes
    grpc_max_receive_message_length
       Max grpc receive message size in bytes
    names
       Column names
    kwargs

    Returns
    -------
       SeldonClientCombine

    """
    if datas is None:
        datas = []
        for n in range(ndatas):
            data = np.random.rand(*shape)
            datas.append(data)
    msgs = []
    for data in datas:
        if isinstance(data, (bytes, bytearray)):
            msgs.append(prediction_pb2.SeldonMessage(binData=data))
        elif isinstance(data, str):
            msgs.append(prediction_pb2.SeldonMessage(strData=data))
        elif isinstance(data, any_pb2.Any):
            msgs.append(prediction_pb2.SeldonMessage(customData=data))
        else:
            datadef = array_to_grpc_datadef(payload_type, data, names=names)
            msgs.append(prediction_pb2.SeldonMessage(data=datadef))
    request = prediction_pb2.SeldonMessageList(seldonMessages=msgs)
    try:
        channel = grpc.insecure_channel(
            microservice_endpoint,
            options=[
                ("grpc.max_send_message_length", grpc_max_send_message_length),
                ("grpc.max_receive_message_length", grpc_max_receive_message_length),
            ],
        )
        stub = prediction_pb2_grpc.GenericStub(channel)
        response = stub.Aggregate(request=request)
        channel.close()
        return SeldonClientCombine(request, response, True, "")
    except Exception as e:
        return SeldonClientCombine(request, None, False, str(e))


def microservice_api_grpc_feedback(
    prediction_request: prediction_pb2.SeldonMessage = None,
    prediction_response: prediction_pb2.SeldonMessage = None,
    reward: float = 0,
    microservice_endpoint: str = None,
    grpc_max_send_message_length: int = 4 * 1024 * 1024,
    grpc_max_receive_message_length: int = 4 * 1024 * 1024,
    **kwargs,
) -> SeldonClientFeedback:
    """
    Call Seldon gRPC

    Parameters
    ----------
    prediction_request
       Previous prediction request
    prediction_response
       Previous prediction response
    reward
       A reward to send in feedback
    microservice_endpoint
       Running microservice endpoint
    kwargs

    Returns
    -------

    """
    request = prediction_pb2.Feedback(
        request=prediction_request, response=prediction_response, reward=reward
    )
    try:
        channel = grpc.insecure_channel(
            microservice_endpoint,
            options=[
                ("grpc.max_send_message_length", grpc_max_send_message_length),
                ("grpc.max_receive_message_length", grpc_max_receive_message_length),
            ],
        )
        stub = prediction_pb2_grpc.GenericStub(channel)
        response = stub.SendFeedback(request=request)
        channel.close()
        return SeldonClientFeedback(request, response, True, "")
    except Exception as e:
        return SeldonClientFeedback(request, None, False, str(e))


#
# External API
#


def rest_predict_seldon(
    namespace: str = None,
    gateway_endpoint: str = "localhost:8002",
    shape: Tuple = (1, 1),
    data: object = None,
    payload_type: str = "tensor",
    bin_data: Union[bytes, bytearray] = None,
    str_data: str = None,
    json_data: Union[str, List, Dict] = None,
    names: Iterable[str] = None,
    client_return_type: str = "proto",
    raw_data: Dict = None,
    meta: Dict = {},
    **kwargs,
) -> SeldonClientPrediction:
    """
    Call Seldon API Gateway using REST

    Parameters
    ----------
    namespace
       k8s namespace of running deployment
    shape
       Shape of endpoint
    data
       Data to send
    payload_type
       payload - tensor, ndarray or tftensor
    bin_data
       Binary data to send
    str_data
       String data to send
    json_data
        JSON data to send
    names
       column names
    client_return_type
        the return type of all functions can be either dict or proto
    raw_data
        Raw payload (dictionary) given by the user
    meta
        Custom meta data map, supplied as tags
    kwargs

    Returns
    -------
       Seldon Client Prediction

    """
    metaKV = prediction_pb2.Meta()
    metaJson = {"tags": meta}
    json_format.ParseDict(metaJson, metaKV)
    if raw_data:
        request = json_to_seldon_message(raw_data)
        payload = raw_data
    else:
        if bin_data is not None:
            request = prediction_pb2.SeldonMessage(binData=bin_data, meta=metaKV)
        elif str_data is not None:
            request = prediction_pb2.SeldonMessage(strData=str_data, meta=metaKV)
        elif json_data is not None:
            request = json_to_seldon_message({"jsonData": json_data})
        else:
            if data is None:
                data = np.random.rand(*shape)
            datadef = array_to_grpc_datadef(payload_type, data, names=names)
            request = prediction_pb2.SeldonMessage(data=datadef, meta=metaKV)
        payload = seldon_message_to_json(request)

    response_raw = requests.post(
        "http://" + gateway_endpoint + "/api/v1.0/predictions", json=payload
    )
    if response_raw.status_code == 200:
        success = True
        msg = ""
    else:
        success = False
        msg = str(response_raw.status_code) + ":" + response_raw.reason
    try:
        if len(response_raw.text) > 0:
            try:
                if client_return_type == "proto":
                    response = json_to_seldon_message(response_raw.json())
                elif client_return_type == "dict":
                    response = response_raw.json()
                else:
                    SeldonClientException("Invalid client_return_type")
            except:
                response = None
        else:
            response = None
        return SeldonClientPrediction(request, response, success, msg)
    except Exception as e:
        return SeldonClientPrediction(request, None, False, str(e))


def grpc_predict_seldon(
    namespace: str = None,
    gateway_endpoint: str = "localhost:8004",
    shape: Tuple[int, int] = (1, 1),
    data: np.ndarray = None,
    payload_type: str = "tensor",
    bin_data: Union[bytes, bytearray] = None,
    str_data: str = None,
    json_data: Union[str, List, Dict] = None,
    custom_data: any_pb2.Any = None,
    grpc_max_send_message_length: int = 4 * 1024 * 1024,
    grpc_max_receive_message_length: int = 4 * 1024 * 1024,
    names: Iterable[str] = None,
    client_return_type: str = "proto",
    raw_data: Dict = None,
    **kwargs,
) -> SeldonClientPrediction:
    """
    Call Seldon gRPC API Gateway endpoint

    Parameters
    ----------
    namespace
       k8s namespace of running deployment
    shape
       Shape of endpoint
    data
       Data to send
    payload_type
       payload - tensor, ndarray or tftensor
    bin_data
       Binary data to send
    str_data
       String data to send
    json_data
       JSON data to send
    custom_data
       Custom data to send
    grpc_max_send_message_length
       Max grpc send message size in bytes
    grpc_max_receive_message_length
       Max grpc receive message size in bytes
    names
       Column names
    client_return_type
        the return type of all functions can be either dict or proto
    raw_data
        Raw payload (dictionary or proto) given by the user
    kwargs

    Returns
    -------
       A SeldonMessage proto

    """
    if isinstance(raw_data, prediction_pb2.SeldonMessage):
        request = raw_data
    elif raw_data:
        request = json_to_seldon_message(raw_data)
    else:
        if bin_data is not None:
            request = prediction_pb2.SeldonMessage(binData=bin_data)
        elif str_data is not None:
            request = prediction_pb2.SeldonMessage(strData=str_data)
        elif json_data is not None:
            request = json_to_seldon_message({"jsonData": json_data})
        elif custom_data is not None:
            request = prediction_pb2.SeldonMessage(customData=custom_data)
        else:
            if data is None:
                data = np.random.rand(*shape)
            datadef = array_to_grpc_datadef(payload_type, data, names=names)
            request = prediction_pb2.SeldonMessage(data=datadef)

    channel = grpc.insecure_channel(
        gateway_endpoint,
        options=[
            ("grpc.max_send_message_length", grpc_max_send_message_length),
            ("grpc.max_receive_message_length", grpc_max_receive_message_length),
        ],
    )
    stub = prediction_pb2_grpc.SeldonStub(channel)
    try:
        response = stub.Predict(request=request)
        channel.close()
        if client_return_type == "dict":
            request = seldon_message_to_json(request)
            response = seldon_message_to_json(response)
        elif client_return_type != "proto":
            raise SeldonClientException("Invalid client_return_type")
        return SeldonClientPrediction(request, response, True, "")
    except Exception as e:
        channel.close()
        return SeldonClientPrediction(request, None, False, str(e))


def rest_predict_gateway(
    deployment_name: str,
    namespace: str = None,
    gateway_endpoint: str = "localhost:8003",
    shape: Tuple[int, int] = (1, 1),
    data: np.ndarray = None,
    headers: Dict = None,
    gateway_prefix: str = None,
    payload_type: str = "tensor",
    bin_data: Union[bytes, bytearray] = None,
    str_data: str = None,
    json_data: Union[str, Dict, List] = None,
    names: Iterable[str] = None,
    call_credentials: SeldonCallCredentials = None,
    channel_credentials: SeldonChannelCredentials = None,
    http_path: str = None,
    meta: Dict = {},
    client_return_type: str = "proto",
    raw_data: Dict = None,
    ssl: bool = None,
    **kwargs,
) -> SeldonClientPrediction:
    """
    REST request to Gateway Ingress

    Parameters
    ----------
    deployment_name
       The name of the Seldon Deployment
    namespace
       k8s namespace of running deployment
    gateway_endpoint
       The host:port of gateway
    shape
       The shape of the data to send
    data
       The numpy data to send
    headers
       Headers to add to request
    gateway_prefix
       The prefix path to add to the request
    payload_type
       payload - tensor, ndarray or tftensor
    bin_data
       Binary data to send
    str_data
       String data to send
    json_data
       JSON data to send as str, dict or list
    names
       Column names
    call_credentials
       Call credentials - see SeldonCallCredentials
    channel_credentials
       Channel credentials - see SeldonChannelCredentials
    http_path
       Custom http path
    meta
       Custom meta map
    client_return_type
        the return type of all functions can be either dict or proto
    raw_data
        Raw payload (dictionary) given by the user

    Returns
    -------
       A requests Response object

    """
    # Create meta data
    metaKV = prediction_pb2.Meta()
    metaJson = {"tags": meta}
    json_format.ParseDict(metaJson, metaKV)
    if raw_data is not None:
        request = json_to_seldon_message(raw_data)
        payload = raw_data
    else:
        if bin_data is not None:
            request = prediction_pb2.SeldonMessage(binData=bin_data, meta=metaKV)
        elif str_data is not None:
            request = prediction_pb2.SeldonMessage(strData=str_data, meta=metaKV)
        elif json_data is not None:
            request = json_to_seldon_message({"jsonData": json_data})
        else:
            if data is None:
                data = np.random.rand(*shape)
            datadef = array_to_grpc_datadef(payload_type, data, names=names)
            request = prediction_pb2.SeldonMessage(data=datadef, meta=metaKV)
        payload = seldon_message_to_json(request)

    if headers is not None:
        req_headers = headers.copy()
    else:
        req_headers = {}
    if call_credentials is None or ssl is False:
        scheme = "http"
    else:
        scheme = "https"
    if call_credentials is not None:
        if call_credentials.token is not None:
            req_headers["X-Auth-Token"] = call_credentials.token
    if http_path is not None:
        url = (
            scheme
            + "://"
            + gateway_endpoint
            + "/seldon/"
            + namespace
            + "/"
            + deployment_name
            + http_path
        )
    else:
        if gateway_prefix is None:
            if namespace is None:
                url = (
                    scheme
                    + "://"
                    + gateway_endpoint
                    + "/seldon/"
                    + deployment_name
                    + "/api/v1.0/predictions"
                )
            else:
                url = (
                    scheme
                    + "://"
                    + gateway_endpoint
                    + "/seldon/"
                    + namespace
                    + "/"
                    + deployment_name
                    + "/api/v1.0/predictions"
                )
        else:
            url = (
                scheme
                + "://"
                + gateway_endpoint
                + gateway_prefix
                + "/api/v1.0/predictions"
            )
    verify = True
    cert = None
    if channel_credentials is not None:
        if channel_credentials.certificate_chain_file is not None:
            verify = channel_credentials.certificate_chain_file
        else:
            verify = channel_credentials.verify
        if channel_credentials.private_key_file is not None:
            cert = (
                channel_credentials.root_certificates_file,
                channel_credentials.private_key_file,
            )
    logger.debug("URL is " + url)
    response_raw = requests.post(
        url, json=payload, headers=req_headers, verify=verify, cert=cert
    )
    if response_raw.status_code == 200:
        success = True
        msg = ""
    else:
        success = False
        msg = str(response_raw.status_code) + ":" + response_raw.reason
    try:
        if len(response_raw.text) > 0:
            try:
                logger.debug("Raw response: %s", response_raw.text)
                if client_return_type == "proto":
                    response = json_to_seldon_message(response_raw.json())
                elif client_return_type == "dict":
                    response = response_raw.json()
                else:
                    raise SeldonClientException("Invalid client_return_type")
            except:
                response = None
        else:
            response = None
        return SeldonClientPrediction(request, response, success, msg)
    except Exception as e:
        return SeldonClientPrediction(request, None, False, str(e))


def explain_predict_gateway(
    deployment_name: str,
    namespace: str = None,
    gateway_endpoint: str = "localhost:8003",
    gateway: str = None,
    transport: str = "rest",
    shape: Tuple[int, int] = (1, 1),
    data: np.ndarray = None,
    headers: Dict = None,
    gateway_prefix: str = None,
    payload_type: str = "tensor",
    bin_data: Union[bytes, bytearray] = None,
    str_data: str = None,
    json_data: Union[str, List, Dict] = None,
    names: Iterable[str] = None,
    call_credentials: SeldonCallCredentials = None,
    channel_credentials: SeldonChannelCredentials = None,
    http_path: str = None,
    client_return_type: str = "dict",
    predictor: str = None,
    ssl: bool = None,
    **kwargs,
) -> SeldonClientPrediction:
    """
    REST explain request to Gateway Ingress

    Parameters
    ----------
    deployment_name
       The name of the Seldon Deployment
    namespace
       k8s namespace of running deployment
    gateway_endpoint
       The host:port of gateway
    gateway
       The type of gateway which can be seldon or ambassador/istio
    transport
       The type of transport, in this case only rest is supported
    shape
       The shape of the data to send
    data
       The numpy data to send
    headers
       Headers to add to request
    gateway_prefix
       The prefix path to add to the request
    payload_type
       payload - tensor, ndarray or tftensor
    bin_data
       Binary data to send
    str_data
       String data to send
    json_data
       JSON data to send
    names
       Column names
    call_credentials
       Call credentials - see SeldonCallCredentials
    channel_credentials
       Channel credentials - see SeldonChannelCredentials
    http_path
       Custom http path
    client_return_type
        the return type of all functions can be either dict or proto

    Returns
    -------
       A JSON Dict

    """
    if transport != "rest":
        raise SeldonClientException("Only supported transport is REST for explanations")

    if bin_data is not None:
        request = prediction_pb2.SeldonMessage(binData=bin_data)
    elif str_data is not None:
        request = prediction_pb2.SeldonMessage(strData=str_data)
    elif json_data is not None:
        request = json_to_seldon_message({"jsonData": json_data})
    else:
        if data is None:
            data = np.random.rand(*shape)
        datadef = array_to_grpc_datadef(payload_type, data, names=names)
        request = prediction_pb2.SeldonMessage(data=datadef)
    payload = seldon_message_to_json(request)

    if not headers is None:
        req_headers = headers.copy()
    else:
        req_headers = {}
    if channel_credentials is None or ssl is False:
        scheme = "http"
    else:
        scheme = "https"
    if not call_credentials is None:
        if not call_credentials.token is None:
            req_headers["X-Auth-Token"] = call_credentials.token
    if http_path is not None:
        url = (
            scheme
            + "://"
            + gateway_endpoint
            + "/seldon/"
            + namespace
            + "/"
            + deployment_name
            + "-explainer"
            + "/"
            + predictor
            + http_path
        )
    elif gateway == "seldon":
        url = scheme + "://" + gateway_endpoint + "/api/v1.0/explain"
    else:
        if not predictor:
            raise SeldonClientException(
                "Predictor parameter must be provided to talk through explainer via gateway"
            )

        if gateway_prefix is None:
            if namespace is None:
                url = (
                    scheme
                    + "://"
                    + gateway_endpoint
                    + "/seldon/"
                    + deployment_name
                    + "-explainer"
                    + "/"
                    + predictor
                    + "/api/v1.0/explain"
                )
            else:
                url = (
                    scheme
                    + "://"
                    + gateway_endpoint
                    + "/seldon/"
                    + namespace
                    + "/"
                    + deployment_name
                    + "-explainer"
                    + "/"
                    + predictor
                    + "/api/v1.0/explain"
                )
        else:
            url = (
                scheme + "://" + gateway_endpoint + gateway_prefix + "/api/v1.0/explain"
            )
    verify = True
    cert = None
    if not channel_credentials is None:
        if not channel_credentials.certificate_chain_file is None:
            verify = channel_credentials.certificate_chain_file
        else:
            verify = channel_credentials.verify
        if not channel_credentials.private_key_file is None:
            cert = (
                channel_credentials.root_certificates_file,
                channel_credentials.private_key_file,
            )
    logger.debug("URL is " + url)
    response_raw = requests.post(
        url, json=payload, headers=req_headers, verify=verify, cert=cert
    )
    if response_raw.status_code == 200:
        if client_return_type == "dict":
            ret_request = payload
            ret_response = response_raw.json()
        else:
            raise SeldonClientException("Invalid client_return_type")
        return SeldonClientPrediction(ret_request, ret_response, True, "")
    else:
        return SeldonClientPrediction(
            payload,
            response_raw,
            False,
            f"Unsuccessful request with status code: {response_raw.status_code}",
        )


def grpc_predict_gateway(
    deployment_name: str,
    namespace: str = None,
    gateway_endpoint: str = "localhost:8003",
    shape: Tuple[int, int] = (1, 1),
    data: np.ndarray = None,
    headers: Dict = None,
    payload_type: str = "tensor",
    bin_data: Union[bytes, bytearray] = None,
    str_data: str = None,
    json_data: Union[str, List, Dict] = None,
    custom_data: any_pb2.Any = None,
    grpc_max_send_message_length: int = 4 * 1024 * 1024,
    grpc_max_receive_message_length: int = 4 * 1024 * 1024,
    names: Iterable[str] = None,
    call_credentials: SeldonCallCredentials = None,
    channel_credentials: SeldonChannelCredentials = None,
    meta: Dict = {},
    client_return_type: str = "proto",
    raw_data: Dict = None,
    ssl: bool = None,
    **kwargs,
) -> SeldonClientPrediction:
    """
    gRPC request to Gateway Ingress

    Parameters
    ----------
    deployment_name
       Deployment name of Seldon Deployment
    namespace
       The namespace the Seldon Deployment is running in
    gateway_endpoint
       The endpoint for gateway
    shape
       The shape of the data
    data
       The numpy array data to send
    headers
      A Dict of key value pairs to add to gRPC HTTP Headers
    payload_type
       payload - tensor, ndarray or tftensor
    bin_data
       Binary data to send
    str_data
       String data to send
    json_data
       JSON data to send
    custom_data
       Custom data to send
    grpc_max_send_message_length
       Max grpc send message size in bytes
    grpc_max_receive_message_length
       Max grpc receive message size in bytes
    names
       Column names
    call_credentials
       Call credentials - see SeldonCallCredentials
    channel_credentials
       Channel credentials - see SeldonChannelCredentials
    meta
       Custom meta data map, supplied as tags
    client_return_type
        the return type of all functions can be either dict or proto
    raw_data
        Raw payload (dictionary or proto) given by the user

    Returns
    -------
       A SeldonMessage proto response

    """

    # Create meta data
    metaKV = prediction_pb2.Meta()
    metaJson = {"tags": meta}
    json_format.ParseDict(metaJson, metaKV)

    if isinstance(raw_data, prediction_pb2.SeldonMessage):
        request = raw_data
    elif raw_data is not None:
        request = json_to_seldon_message(raw_data)
    else:
        if bin_data is not None:
            request = prediction_pb2.SeldonMessage(binData=bin_data, meta=metaKV)
        elif str_data is not None:
            request = prediction_pb2.SeldonMessage(strData=str_data, meta=metaKV)
        elif json_data is not None:
            request = json_to_seldon_message({"jsonData": json_data})
        elif custom_data is not None:
            request = prediction_pb2.SeldonMessage(customData=custom_data, meta=metaKV)
        else:
            if data is None:
                data = np.random.rand(*shape)
            datadef = array_to_grpc_datadef(payload_type, data, names=names)
            request = prediction_pb2.SeldonMessage(data=datadef, meta=metaKV)
    options = [
        ("grpc.max_send_message_length", grpc_max_send_message_length),
        ("grpc.max_receive_message_length", grpc_max_receive_message_length),
    ]
    if channel_credentials is None:
        channel = grpc.insecure_channel(gateway_endpoint, options)
    else:
        # If one of root cert & cert chain are provided, both must be provided
        #   otherwise there is a null pointer exception in the Go underlying impl
        if (
            channel_credentials.private_key_file
            and channel_credentials.root_certificates_file
            and channel_credentials.certificate_chain_file
        ):
            grpc_channel_credentials = grpc.ssl_channel_credentials(
                root_certificates=open(
                    channel_credentials.root_certificates_file, "rb"
                ).read(),
                private_key=open(channel_credentials.private_key_file, "rb").read(),
                certificate_chain=open(
                    channel_credentials.certificate_chain_file, "rb"
                ).read(),
            )
        # For most usecases only providing the root cert file is enough
        elif channel_credentials.root_certificates_file:
            grpc_channel_credentials = grpc.ssl_channel_credentials(
                root_certificates=open(
                    channel_credentials.root_certificates_file, "rb"
                ).read()
            )
        # This piece also allows for blank SSL Channel credentials in case this is required
        else:
            if ssl is False:
                grpc_channel_credentials = grpc.local_channel_credentials()
            else:
                grpc_channel_credentials = grpc.ssl_channel_credentials()
        if channel_credentials.verify == False:
            # If Verify is set to false then we add the SSL Target Name Override option
            options += [
                ("grpc.ssl_target_name_override", gateway_endpoint.split(":")[0])
            ]

        if not call_credentials is None:
            grpc_call_credentials = grpc.metadata_call_credentials(
                lambda context, callback: callback(
                    (("x-auth-token", call_credentials.token),), None
                )
            )
            credentials = grpc.composite_channel_credentials(
                grpc_channel_credentials, grpc_call_credentials
            )
        else:
            credentials = grpc_channel_credentials
        logger.debug(f"Sending GRPC Request to endpoint: {gateway_endpoint}")
        channel = grpc.secure_channel(gateway_endpoint, credentials, options)
    stub = prediction_pb2_grpc.SeldonStub(channel)
    if namespace is None:
        metadata = [("seldon", deployment_name)]
    else:
        metadata = [("seldon", deployment_name), ("namespace", namespace)]
    if not headers is None:
        for k in headers:
            metadata.append((k, headers[k]))
    try:
        response = stub.Predict(request=request, metadata=metadata)
        channel.close()
        if client_return_type == "dict":
            request = seldon_message_to_json(request)
            response = seldon_message_to_json(response)
        elif client_return_type != "proto":
            raise SeldonClientException("Invalid client_return_type")
        return SeldonClientPrediction(request, response, True, "")
    except Exception as e:
        channel.close()
        return SeldonClientPrediction(request, None, False, str(e))


def rest_feedback_seldon(
    prediction_request: prediction_pb2.SeldonMessage = None,
    prediction_response: prediction_pb2.SeldonMessage = None,
    prediction_truth: prediction_pb2.SeldonMessage = None,
    reward: float = 0,
    namespace: str = None,
    gateway_endpoint: str = "localhost:8002",
    client_return_type: str = "proto",
    raw_request: dict = None,
    **kwargs,
) -> SeldonClientFeedback:
    """
    Send Feedback to Seldon API Gateway using REST

    Parameters
    ----------
    prediction_request
       Previous prediction request
    prediction_response
       Previous prediction response
    reward
       A reward to send in feedback
    namespace
       k8s namespace of running deployment
    client_return_type
        the return type of all functions can be either dict or proto
    kwargs

    Returns
    -------

    """

    if raw_request:
        request = json_to_feedback(raw_request)
        payload = raw_request
    else:
        request = prediction_pb2.Feedback(
            request=prediction_request,
            response=prediction_response,
            reward=reward,
            truth=prediction_truth,
        )
        payload = feedback_to_json(request)

    response_raw = requests.post(
        "http://" + gateway_endpoint + "/api/v1.0/feedback", json=payload
    )
    if response_raw.status_code == 200:
        success = True
        msg = ""
    else:
        success = False
        msg = str(response_raw.status_code) + ":" + response_raw.reason
    try:
        if len(response_raw.text) > 0:
            try:
                if client_return_type == "proto":
                    response = json_to_seldon_message(response_raw.json())
                elif client_return_type == "dict":
                    response = response_raw.json()
                else:
                    raise SeldonClientException("Invalid client_return_type")
            except:
                response = None
        else:
            response = None
        return SeldonClientFeedback(request, response, success, msg)
    except Exception as e:
        return SeldonClientFeedback(request, None, False, str(e))


def grpc_feedback_seldon(
    prediction_request: prediction_pb2.SeldonMessage = None,
    prediction_response: prediction_pb2.SeldonMessage = None,
    prediction_truth: prediction_pb2.SeldonMessage = None,
    reward: float = 0,
    namespace: str = None,
    gateway_endpoint: str = "localhost:8004",
    grpc_max_send_message_length: int = 4 * 1024 * 1024,
    grpc_max_receive_message_length: int = 4 * 1024 * 1024,
    client_return_type: str = "proto",
    raw_request: dict = None,
    **kwargs,
) -> SeldonClientFeedback:
    """
    Send feedback to Seldon API gateway via gRPC

    Parameters
    ----------
    prediction_request
       Previous prediction request
    prediction_response
       Previous prediction response
    reward
       A reward to send in feedback
    namespace
       k8s namespace of running deployment
    grpc_max_send_message_length
       Max grpc send message size in bytes
    grpc_max_receive_message_length
       Max grpc receive message size in bytes
    client_return_type
        the return type of all functions can be either dict or proto
    kwargs

    Returns
    -------

    """

    if isinstance(raw_request, prediction_pb2.Feedback):
        request = raw_request
    elif raw_request:
        request = json_to_feedback(raw_request)
    else:
        request = prediction_pb2.Feedback(
            request=prediction_request,
            response=prediction_response,
            reward=reward,
            truth=prediction_truth,
        )

    channel = grpc.insecure_channel(
        gateway_endpoint,
        options=[
            ("grpc.max_send_message_length", grpc_max_send_message_length),
            ("grpc.max_receive_message_length", grpc_max_receive_message_length),
        ],
    )
    stub = prediction_pb2_grpc.SeldonStub(channel)
    try:
        response = stub.SendFeedback(request=request)
        channel.close()
        if client_return_type == "dict":
            request = seldon_message_to_json(request)
            response = seldon_message_to_json(response)
        elif client_return_type != "proto":
            raise SeldonClientException("Invalid client_return_type")
        return SeldonClientFeedback(request, response, True, "")
    except Exception as e:
        channel.close()
        return SeldonClientFeedback(request, None, False, str(e))


def rest_feedback_gateway(
    prediction_request: prediction_pb2.SeldonMessage = None,
    prediction_response: prediction_pb2.SeldonMessage = None,
    prediction_truth: prediction_pb2.SeldonMessage = None,
    reward: float = 0,
    deployment_name: str = "",
    namespace: str = None,
    gateway_endpoint: str = "localhost:8003",
    headers: Dict = None,
    gateway_prefix: str = None,
    client_return_type: str = "proto",
    raw_request: dict = None,
    **kwargs,
) -> SeldonClientFeedback:
    """
    Send Feedback to Seldon via gateway using REST

    Parameters
    ----------
    prediction_request
       Previous prediction request
    prediction_response
       Previous prediction response
    reward
       A reward to send in feedback
    deployment_name
       The name of the running Seldon deployment
    namespace
       k8s namespace of running deployment
    gateway_endpoint
       The gateway host:port endpoint
    headers
       Headers to add to the request
    gateway_prefix
      The prefix to add to the request path for gateway
    client_return_type
        the return type of all functions can be either dict or proto
    kwargs

    Returns
    -------
      A Seldon Feedback Response

    """
    if raw_request:
        request = json_to_feedback(raw_request)
        payload = raw_request
    else:
        request = prediction_pb2.Feedback(
            request=prediction_request,
            response=prediction_response,
            reward=reward,
            truth=prediction_truth,
        )
        payload = feedback_to_json(request)
    if gateway_prefix is None:
        if namespace is None:
            response_raw = requests.post(
                "http://"
                + gateway_endpoint
                + "/seldon/"
                + deployment_name
                + "/api/v1.0/feedback",
                json=payload,
                headers=headers,
            )
        else:
            response_raw = requests.post(
                "http://"
                + gateway_endpoint
                + "/seldon/"
                + namespace
                + "/"
                + deployment_name
                + "/api/v1.0/feedback",
                json=payload,
                headers=headers,
            )
    else:
        response_raw = requests.post(
            "http://" + gateway_endpoint + gateway_prefix + "/api/v1.0/feedback",
            json=payload,
            headers=headers,
        )

    if response_raw.status_code == 200:
        success = True
        msg = ""
    else:
        success = False
        msg = str(response_raw.status_code) + ":" + response_raw.reason
    try:
        if len(response_raw.text) > 0:
            try:
                if client_return_type == "proto":
                    response = json_to_seldon_message(response_raw.json())
                elif client_return_type == "dict":
                    response = response_raw.json()
                else:
                    raise SeldonClientException("Invalid client_return_type")
            except:
                response = None
        else:
            response = None
        return SeldonClientFeedback(request, response, success, msg)
    except Exception as e:
        return SeldonClientFeedback(request, None, False, str(e))


def grpc_feedback_gateway(
    prediction_request: prediction_pb2.SeldonMessage = None,
    prediction_response: prediction_pb2.SeldonMessage = None,
    prediction_truth: prediction_pb2.SeldonMessage = None,
    reward: float = 0,
    deployment_name: str = "",
    namespace: str = None,
    gateway_endpoint: str = "localhost:8003",
    headers: Dict = None,
    grpc_max_send_message_length: int = 4 * 1024 * 1024,
    grpc_max_receive_message_length: int = 4 * 1024 * 1024,
    client_return_type: str = "proto",
    raw_request: dict = None,
    **kwargs,
) -> SeldonClientFeedback:
    """

    Parameters
    ----------
    prediction_request
       Previous prediction request
    prediction_response
       Previous prediction response
    reward
       A reward to send in feedback
    deployment_name
       The name of the running Seldon deployment
    namespace
       k8s namespace of running deployment
    gateway_endpoint
       The gateway host:port endpoint
    headers
       Headers to add to the request
    grpc_max_send_message_length
       Max grpc send message size in bytes
    grpc_max_receive_message_length
       Max grpc receive message size in bytes
    client_return_type
        the return type of all functions can be either dict or proto
    kwargs

    Returns
    -------

    """
    if isinstance(raw_request, prediction_pb2.Feedback):
        request = raw_request
    elif raw_request:
        request = json_to_feedback(raw_request)
    else:
        request = prediction_pb2.Feedback(
            request=prediction_request,
            response=prediction_response,
            reward=reward,
            truth=prediction_truth,
        )
    channel = grpc.insecure_channel(
        gateway_endpoint,
        options=[
            ("grpc.max_send_message_length", grpc_max_send_message_length),
            ("grpc.max_receive_message_length", grpc_max_receive_message_length),
        ],
    )
    stub = prediction_pb2_grpc.SeldonStub(channel)
    if namespace is None:
        metadata = [("seldon", deployment_name)]
    else:
        metadata = [("seldon", deployment_name), ("namespace", namespace)]
    if not headers is None:
        for k in headers:
            metadata.append((k, headers[k]))
    try:
        response = stub.SendFeedback(request=request, metadata=metadata)
        channel.close()
        if client_return_type == "dict":
            request = seldon_message_to_json(request)
            response = seldon_message_to_json(response)
        elif client_return_type != "proto":
            raise SeldonClientException("Invalid client_return_type")
        return SeldonClientFeedback(request, response, True, "")
    except Exception as e:
        channel.close()
        return SeldonClientFeedback(request, None, False, str(e))



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import collections.abc
import google.protobuf.any_pb2
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import google.protobuf.struct_pb2
import sys
import tensorflow.core.framework.tensor_pb2
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class SeldonMessage(google.protobuf.message.Message):
    """[START Messages]"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    STATUS_FIELD_NUMBER: builtins.int
    META_FIELD_NUMBER: builtins.int
    DATA_FIELD_NUMBER: builtins.int
    BINDATA_FIELD_NUMBER: builtins.int
    STRDATA_FIELD_NUMBER: builtins.int
    JSONDATA_FIELD_NUMBER: builtins.int
    CUSTOMDATA_FIELD_NUMBER: builtins.int
    @property
    def status(self) -> global___Status: ...
    @property
    def meta(self) -> global___Meta: ...
    @property
    def data(self) -> global___DefaultData: ...
    binData: builtins.bytes
    strData: builtins.str
    @property
    def jsonData(self) -> google.protobuf.struct_pb2.Value: ...
    @property
    def customData(self) -> google.protobuf.any_pb2.Any: ...
    def __init__(
        self,
        *,
        status: global___Status | None = ...,
        meta: global___Meta | None = ...,
        data: global___DefaultData | None = ...,
        binData: builtins.bytes = ...,
        strData: builtins.str = ...,
        jsonData: google.protobuf.struct_pb2.Value | None = ...,
        customData: google.protobuf.any_pb2.Any | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["binData", b"binData", "customData", b"customData", "data", b"data", "data_oneof", b"data_oneof", "jsonData", b"jsonData", "meta", b"meta", "status", b"status", "strData", b"strData"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["binData", b"binData", "customData", b"customData", "data", b"data", "data_oneof", b"data_oneof", "jsonData", b"jsonData", "meta", b"meta", "status", b"status", "strData", b"strData"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["data_oneof", b"data_oneof"]) -> typing_extensions.Literal["data", "binData", "strData", "jsonData", "customData"] | None: ...

global___SeldonMessage = SeldonMessage

@typing_extensions.final
class DefaultData(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAMES_FIELD_NUMBER: builtins.int
    TENSOR_FIELD_NUMBER: builtins.int
    NDARRAY_FIELD_NUMBER: builtins.int
    TFTENSOR_FIELD_NUMBER: builtins.int
    @property
    def names(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    @property
    def tensor(self) -> global___Tensor: ...
    @property
    def ndarray(self) -> google.protobuf.struct_pb2.ListValue: ...
    @property
    def tftensor(self) -> tensorflow.core.framework.tensor_pb2.TensorProto: ...
    def __init__(
        self,
        *,
        names: collections.abc.Iterable[builtins.str] | None = ...,
        tensor: global___Tensor | None = ...,
        ndarray: google.protobuf.struct_pb2.ListValue | None = ...,
        tftensor: tensorflow.core.framework.tensor_pb2.TensorProto | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["data_oneof", b"data_oneof", "ndarray", b"ndarray", "tensor", b"tensor", "tftensor", b"tftensor"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["data_oneof", b"data_oneof", "names", b"names", "ndarray", b"ndarray", "tensor", b"tensor", "tftensor", b"tftensor"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["data_oneof", b"data_oneof"]) -> typing_extensions.Literal["tensor", "ndarray", "tftensor"] | None: ...

global___DefaultData = DefaultData

@typing_extensions.final
class Tensor(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SHAPE_FIELD_NUMBER: builtins.int
    VALUES_FIELD_NUMBER: builtins.int
    @property
    def shape(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    @property
    def values(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.float]: ...
    def __init__(
        self,
        *,
        shape: collections.abc.Iterable[builtins.int] | None = ...,
        values: collections.abc.Iterable[builtins.float] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["shape", b"shape", "values", b"values"]) -> None: ...

global___Tensor = Tensor

@typing_extensions.final
class Meta(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing_extensions.final
    class TagsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        @property
        def value(self) -> google.protobuf.struct_pb2.Value: ...
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: google.protobuf.struct_pb2.Value | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["value", b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    @typing_extensions.final
    class RoutingEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.int
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.int = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    @typing_extensions.final
    class RequestPathEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.str
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    PUID_FIELD_NUMBER: builtins.int
    TAGS_FIELD_NUMBER: builtins.int
    ROUTING_FIELD_NUMBER: builtins.int
    REQUESTPATH_FIELD_NUMBER: builtins.int
    METRICS_FIELD_NUMBER: builtins.int
    puid: builtins.str
    @property
    def tags(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, google.protobuf.struct_pb2.Value]: ...
    @property
    def routing(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.int]: ...
    @property
    def requestPath(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]: ...
    @property
    def metrics(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Metric]: ...
    def __init__(
        self,
        *,
        puid: builtins.str = ...,
        tags: collections.abc.Mapping[builtins.str, google.protobuf.struct_pb2.Value] | None = ...,
        routing: collections.abc.Mapping[builtins.str, builtins.int] | None = ...,
        requestPath: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
        metrics: collections.abc.Iterable[global___Metric] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["metrics", b"metrics", "puid", b"puid", "requestPath", b"requestPath", "routing", b"routing", "tags", b"tags"]) -> None: ...

global___Meta = Meta

@typing_extensions.final
class Metric(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _MetricType:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _MetricTypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[Metric._MetricType.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        COUNTER: Metric._MetricType.ValueType  # 0
        GAUGE: Metric._MetricType.ValueType  # 1
        TIMER: Metric._MetricType.ValueType  # 2

    class MetricType(_MetricType, metaclass=_MetricTypeEnumTypeWrapper): ...
    COUNTER: Metric.MetricType.ValueType  # 0
    GAUGE: Metric.MetricType.ValueType  # 1
    TIMER: Metric.MetricType.ValueType  # 2

    @typing_extensions.final
    class TagsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.str
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    KEY_FIELD_NUMBER: builtins.int
    TYPE_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    TAGS_FIELD_NUMBER: builtins.int
    key: builtins.str
    type: global___Metric.MetricType.ValueType
    value: builtins.float
    @property
    def tags(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]: ...
    def __init__(
        self,
        *,
        key: builtins.str = ...,
        type: global___Metric.MetricType.ValueType = ...,
        value: builtins.float = ...,
        tags: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "tags", b"tags", "type", b"type", "value", b"value"]) -> None: ...

global___Metric = Metric

@typing_extensions.final
class SeldonMessageList(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SELDONMESSAGES_FIELD_NUMBER: builtins.int
    @property
    def seldonMessages(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___SeldonMessage]: ...
    def __init__(
        self,
        *,
        seldonMessages: collections.abc.Iterable[global___SeldonMessage] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["seldonMessages", b"seldonMessages"]) -> None: ...

global___SeldonMessageList = SeldonMessageList

@typing_extensions.final
class Status(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _StatusFlag:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _StatusFlagEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[Status._StatusFlag.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        SUCCESS: Status._StatusFlag.ValueType  # 0
        FAILURE: Status._StatusFlag.ValueType  # 1

    class StatusFlag(_StatusFlag, metaclass=_StatusFlagEnumTypeWrapper): ...
    SUCCESS: Status.StatusFlag.ValueType  # 0
    FAILURE: Status.StatusFlag.ValueType  # 1

    CODE_FIELD_NUMBER: builtins.int
    INFO_FIELD_NUMBER: builtins.int
    REASON_FIELD_NUMBER: builtins.int
    STATUS_FIELD_NUMBER: builtins.int
    code: builtins.int
    info: builtins.str
    reason: builtins.str
    status: global___Status.StatusFlag.ValueType
    def __init__(
        self,
        *,
        code: builtins.int = ...,
        info: builtins.str = ...,
        reason: builtins.str = ...,
        status: global___Status.StatusFlag.ValueType = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["code", b"code", "info", b"info", "reason", b"reason", "status", b"status"]) -> None: ...

global___Status = Status

@typing_extensions.final
class Feedback(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    REQUEST_FIELD_NUMBER: builtins.int
    RESPONSE_FIELD_NUMBER: builtins.int
    REWARD_FIELD_NUMBER: builtins.int
    TRUTH_FIELD_NUMBER: builtins.int
    @property
    def request(self) -> global___SeldonMessage: ...
    @property
    def response(self) -> global___SeldonMessage: ...
    reward: builtins.float
    @property
    def truth(self) -> global___SeldonMessage: ...
    def __init__(
        self,
        *,
        request: global___SeldonMessage | None = ...,
        response: global___SeldonMessage | None = ...,
        reward: builtins.float = ...,
        truth: global___SeldonMessage | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["request", b"request", "response", b"response", "truth", b"truth"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["request", b"request", "response", b"response", "reward", b"reward", "truth", b"truth"]) -> None: ...

global___Feedback = Feedback

@typing_extensions.final
class RequestResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    REQUEST_FIELD_NUMBER: builtins.int
    RESPONSE_FIELD_NUMBER: builtins.int
    @property
    def request(self) -> global___SeldonMessage: ...
    @property
    def response(self) -> global___SeldonMessage: ...
    def __init__(
        self,
        *,
        request: global___SeldonMessage | None = ...,
        response: global___SeldonMessage | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["request", b"request", "response", b"response"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["request", b"request", "response", b"response"]) -> None: ...

global___RequestResponse = RequestResponse

@typing_extensions.final
class SeldonModelMetadataRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    name: builtins.str
    def __init__(
        self,
        *,
        name: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["name", b"name"]) -> None: ...

global___SeldonModelMetadataRequest = SeldonModelMetadataRequest

@typing_extensions.final
class SeldonMessageMetadata(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MESSAGETYPE_FIELD_NUMBER: builtins.int
    SCHEMA_FIELD_NUMBER: builtins.int
    NAME_FIELD_NUMBER: builtins.int
    DATATYPE_FIELD_NUMBER: builtins.int
    SHAPE_FIELD_NUMBER: builtins.int
    messagetype: builtins.str
    """SeldonMessage Metadata fields"""
    @property
    def schema(self) -> google.protobuf.struct_pb2.Value: ...
    name: builtins.str
    """KFserving tesnor metadata fields"""
    datatype: builtins.str
    @property
    def shape(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    def __init__(
        self,
        *,
        messagetype: builtins.str = ...,
        schema: google.protobuf.struct_pb2.Value | None = ...,
        name: builtins.str = ...,
        datatype: builtins.str = ...,
        shape: collections.abc.Iterable[builtins.int] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["schema", b"schema"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["datatype", b"datatype", "messagetype", b"messagetype", "name", b"name", "schema", b"schema", "shape", b"shape"]) -> None: ...

global___SeldonMessageMetadata = SeldonMessageMetadata

@typing_extensions.final
class SeldonModelMetadata(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing_extensions.final
    class CustomEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.str
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    NAME_FIELD_NUMBER: builtins.int
    VERSIONS_FIELD_NUMBER: builtins.int
    PLATFORM_FIELD_NUMBER: builtins.int
    INPUTS_FIELD_NUMBER: builtins.int
    OUTPUTS_FIELD_NUMBER: builtins.int
    CUSTOM_FIELD_NUMBER: builtins.int
    name: builtins.str
    @property
    def versions(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    platform: builtins.str
    @property
    def inputs(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___SeldonMessageMetadata]: ...
    @property
    def outputs(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___SeldonMessageMetadata]: ...
    @property
    def custom(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]: ...
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        versions: collections.abc.Iterable[builtins.str] | None = ...,
        platform: builtins.str = ...,
        inputs: collections.abc.Iterable[global___SeldonMessageMetadata] | None = ...,
        outputs: collections.abc.Iterable[global___SeldonMessageMetadata] | None = ...,
        custom: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["custom", b"custom", "inputs", b"inputs", "name", b"name", "outputs", b"outputs", "platform", b"platform", "versions", b"versions"]) -> None: ...

global___SeldonModelMetadata = SeldonModelMetadata

@typing_extensions.final
class SeldonGraphMetadata(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing_extensions.final
    class ModelsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        @property
        def value(self) -> global___SeldonModelMetadata: ...
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: global___SeldonModelMetadata | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["value", b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]) -> None: ...

    NAME_FIELD_NUMBER: builtins.int
    MODELS_FIELD_NUMBER: builtins.int
    INPUTS_FIELD_NUMBER: builtins.int
    OUTPUTS_FIELD_NUMBER: builtins.int
    name: builtins.str
    @property
    def models(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, global___SeldonModelMetadata]: ...
    @property
    def inputs(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___SeldonMessageMetadata]: ...
    @property
    def outputs(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___SeldonMessageMetadata]: ...
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        models: collections.abc.Mapping[builtins.str, global___SeldonModelMetadata] | None = ...,
        inputs: collections.abc.Iterable[global___SeldonMessageMetadata] | None = ...,
        outputs: collections.abc.Iterable[global___SeldonMessageMetadata] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["inputs", b"inputs", "models", b"models", "name", b"name", "outputs", b"outputs"]) -> None: ...

global___SeldonGraphMetadata = SeldonGraphMetadata



"""Contains methods to generate a JSON file for Seldon API integration testing."""

import os
from typing import List, Optional, Union

import numpy as np
import pandas as pd

RANGE_INTEGER_MIN = 0
RANGE_INTEGER_MAX = 1
RANGE_FLOAT_MIN = 0.0
RANGE_FLOAT_MAX = 1.0


def _column_range(col: pd.Series) -> Optional[List]:
    """
    Calculate minimum and maximum of a column and outputs a list.

    Parameters
    ----------
    col
        Column to inspect.

    Returns
    -------
        Min and max of the column range as a list.
    """
    if col.dtype == np.float:
        if pd.isnull(min(col)):  # This also means that maximum is null
            return [RANGE_FLOAT_MIN, RANGE_FLOAT_MAX]
        else:
            return [min(col), max(col)]
    elif col.dtype == np.integer:
        if pd.isnull(min(col)):  # This also means that maximum is null
            return [RANGE_INTEGER_MIN, RANGE_INTEGER_MAX]
        else:
            return [min(col), max(col)]
    else:
        return np.NaN


def _column_values(column: pd.Series) -> Union[List, float]:
    """
    Create a list of unique values for categorical variables.

    Parameters
    ----------
    column
        Column to inspect.

    Returns
    -------
        List of unique values for categorical variables
    """
    if column.dtype != np.number:
        return column.unique().tolist()
    else:
        return np.NaN


def create_seldon_api_testing_file(
    data: pd.DataFrame, target: str, output_path: str
) -> bool:
    """
    Create a JSON file for Seldon API testing.

    Parameters
    ----------
    data
        Pandas DataFrame used as a recipe for the json file.
    target
        Name of the target column.
    output_path
        Path of output file.

    Returns
    -------
        True if file correctly generated.
    """

    # create a Data frame in the form of JSON object
    df_for_json = pd.DataFrame(data=data.columns.values, columns=["name"])
    df_for_json["dtype"] = np.where(
        data.dtypes == np.float,
        "FLOAT",
        np.where(data.dtypes == np.int, "INTEGER", np.NaN),
    )
    df_for_json["ftype"] = np.where(
        data.dtypes == np.number, "continuous", "categorical"
    )
    ranges = [_column_range(data[column_name]) for column_name in data.columns.values]
    values = [_column_values(data[column_name]) for column_name in data.columns.values]
    df_for_json["range"] = ranges
    df_for_json["values"] = values
    # Split the target
    df_for_json_target = df_for_json[df_for_json.name == target]
    df_for_json_features = df_for_json[df_for_json.name != target]

    # Convert data frames to JSON with a trick that removes records with NaNs
    json_features_df = df_for_json_features.T.apply(
        lambda row: row[~row.isnull()].to_json()
    )
    json_features = f'[{",".join(json_features_df)}]'
    json_target_df = df_for_json_target.T.apply(
        lambda row: row[~row.isnull()].to_json()
    )
    json_target = f'[{",".join(json_target_df)}]'
    json_combined = f'{{"features": {json_features}, "targets": {json_target}}}'

    with open(output_path, "w+") as output_file:
        output_file.write(str(json_combined))
    return os.path.exists(output_path)



import logging
import textwrap

logger = logging.getLogger(__name__)

# Variables to check if certain extra dependencies are included or
# not
_TF_PRESENT = False

try:
    #  Fix for https://github.com/SeldonIO/seldon-core/issues/1076
    #
    #  If we do `import tensorflow` and there is a folder on the current path
    #  also named `tensorflow`, it may give a false positive even if the folder
    #  is not a Python package. This happens because, since PEP 420 got
    #  introduced in Python 3.3, folders without `__init__.py` can still get
    #  imported as namespaces.
    #
    #  To avoid this and make the check more robust, we test the presence of
    #  the `make_ndarray` method inside the `tensorflow` import.
    from tensorflow import make_ndarray  # noqa: F401

    _TF_PRESENT = True
except ImportError:
    _TF_PRESENT = False
    notice = textwrap.dedent(
        """
        Tensorflow is not installed.
        If you want to use `tftensor` and Tensorflow's data types
        install `tensorflow` or install `seldon_core` as

            $ pip install seldon_core[tensorflow]

        or

            $ pip install seldon_core[all]
        """
    )
    logger.info(notice)



import logging
import os
from concurrent import futures

import grpc
from flask import Flask, Response, request, send_from_directory
from flask_cors import CORS
from grpc_reflection.v1alpha import reflection

import seldon_core.seldon_methods
from seldon_core.env_utils import get_model_name
from seldon_core.flask_utils import (
    ANNOTATION_GRPC_MAX_MSG_SIZE,
    SeldonMicroserviceException,
    get_request,
    jsonify,
)
from seldon_core.proto import prediction_pb2, prediction_pb2_grpc
from seldon_core.utils import (
    getenv_as_bool,
    json_to_feedback,
    json_to_seldon_model_metadata,
    seldon_message_to_json,
)

logger = logging.getLogger(__name__)

PRED_UNIT_ID = get_model_name("0")
METRICS_ENDPOINT = os.environ.get("PREDICTIVE_UNIT_METRICS_ENDPOINT", "/metrics")
PAYLOAD_PASSTHROUGH = getenv_as_bool("PAYLOAD_PASSTHROUGH", default=False)


def get_rest_microservice(user_model, seldon_metrics):
    app = Flask(__name__, static_url_path="")
    CORS(app)

    _set_flask_app_configs(app)

    # dict representing the validated model metadata
    # None value will represent a validation error
    metadata_data = seldon_core.seldon_methods.init_metadata(user_model)

    if hasattr(user_model, "model_error_handler"):
        logger.info("Registering the custom error handler...")
        app.register_blueprint(user_model.model_error_handler)

    @app.errorhandler(Exception)
    def handle_generic_exception(e):
        error = SeldonMicroserviceException(
            message=str(e), status_code=500, reason="MICROSERVICE_INTERNAL_ERROR"
        )
        response = jsonify(error.to_dict())
        logger.error("%s", error.to_dict())
        response.status_code = error.status_code
        return response

    @app.errorhandler(SeldonMicroserviceException)
    def handle_invalid_usage(error):
        response = jsonify(error.to_dict())
        logger.error("%s", error.to_dict())
        response.status_code = error.status_code
        return response

    @app.route("/seldon.json", methods=["GET"])
    def openAPI():
        return send_from_directory("", "openapi/seldon.json")

    @app.route("/predict", methods=["GET", "POST"])
    @app.route("/api/v1.0/predictions", methods=["POST"])
    @app.route("/api/v0.1/predictions", methods=["POST"])
    def Predict():
        requestJson = get_request(skip_decoding=PAYLOAD_PASSTHROUGH)
        logger.debug("REST Request: %s", request)
        response = seldon_core.seldon_methods.predict(
            user_model, requestJson, seldon_metrics
        )

        json_response = jsonify(response, skip_encoding=PAYLOAD_PASSTHROUGH)
        if (
            isinstance(response, dict)
            and "status" in response
            and "code" in response["status"]
        ):
            json_response.status_code = response["status"]["code"]

        logger.debug("REST Response: %s", response)
        return json_response

    @app.route("/send-feedback", methods=["GET", "POST"])
    @app.route("/api/v1.0/feedback", methods=["POST"])
    @app.route("/api/v0.1/feedback", methods=["POST"])
    def SendFeedback():
        requestJson = get_request()
        logger.debug("REST Request: %s", request)
        requestProto = json_to_feedback(requestJson)
        logger.debug("Proto Request: %s", requestProto)
        responseProto = seldon_core.seldon_methods.send_feedback(
            user_model, requestProto, PRED_UNIT_ID, seldon_metrics
        )
        jsonDict = seldon_message_to_json(responseProto)
        return jsonify(jsonDict)

    @app.route("/transform-input", methods=["GET", "POST"])
    def TransformInput():
        requestJson = get_request()
        logger.debug("REST Request: %s", request)
        response = seldon_core.seldon_methods.transform_input(
            user_model, requestJson, seldon_metrics
        )
        logger.debug("REST Response: %s", response)
        return jsonify(response)

    @app.route("/transform-output", methods=["GET", "POST"])
    def TransformOutput():
        requestJson = get_request()
        logger.debug("REST Request: %s", request)
        response = seldon_core.seldon_methods.transform_output(
            user_model, requestJson, seldon_metrics
        )
        logger.debug("REST Response: %s", response)
        return jsonify(response)

    @app.route("/route", methods=["GET", "POST"])
    def Route():
        requestJson = get_request()
        logger.debug("REST Request: %s", request)
        response = seldon_core.seldon_methods.route(
            user_model, requestJson, seldon_metrics
        )
        logger.debug("REST Response: %s", response)
        return jsonify(response)

    @app.route("/aggregate", methods=["GET", "POST"])
    def Aggregate():
        requestJson = get_request()
        logger.debug("REST Request: %s", request)
        response = seldon_core.seldon_methods.aggregate(
            user_model, requestJson, seldon_metrics
        )
        logger.debug("REST Response: %s", response)
        return jsonify(response)

    @app.route("/health/ping", methods=["GET"])
    def HealthPing():
        """
        Lightweight endpoint to check the liveness of the REST endpoint
        """
        return "pong"

    @app.route("/health/status", methods=["GET"])
    @app.route("/api/v1.0/health/status", methods=["GET"])
    def HealthStatus():
        logger.debug("REST Health Status Request")
        response = seldon_core.seldon_methods.health_status(user_model, seldon_metrics)
        logger.debug("REST Health Status Response: %s", response)
        return jsonify(response)

    @app.route("/metadata", methods=["GET"])
    @app.route("/api/v1.0/metadata", methods=["GET"])
    def Metadata():
        if metadata_data is None:
            # None value represents validation error in current implementation
            # if user_model would not define init_metadata than metadata_data
            # would just contain a default values
            raise SeldonMicroserviceException(
                "Model metadata unavailable",
                status_code=500,
                reason="MICROSERVICE_BAD_METADATA",
            )
        logger.debug("REST Metadata Request")
        logger.debug("REST Metadata Response: %s", metadata_data)
        return jsonify(metadata_data)

    return app


def get_metrics_microservice(seldon_metrics):
    app = Flask(__name__, static_url_path="")
    CORS(app)

    _set_flask_app_configs(app)

    @app.route(METRICS_ENDPOINT, methods=["GET"])
    def Metrics():
        logger.debug("REST Metrics Request")
        metrics, mimetype = seldon_metrics.generate_metrics()
        return Response(metrics, mimetype=mimetype)

    return app


def _set_flask_app_configs(app):
    """
    Set the configs for the flask app based on environment variables
    See https://flask.palletsprojects.com/config/#builtin-configuration-values
    :param app:
    :return:
    """
    FLASK_CONFIG_IDENTIFIER = "FLASK_"
    FLASK_CONFIGS_ALLOWED = [
        "DEBUG",
        "EXPLAIN_TEMPLATE_LOADING",
        "JSONIFY_PRETTYPRINT_REGULAR",
        "JSON_AS_ASCII",
        "JSON_SORT_KEYS",
        "PROPAGATE_EXCEPTIONS",
        "PRESERVE_CONTEXT_ON_EXCEPTION",
        "SESSION_COOKIE_HTTPONLY",
        "SESSION_COOKIE_SECURE",
        "SESSION_REFRESH_EACH_REQUEST",
        "TEMPLATES_AUTO_RELOAD",
        "TESTING",
        "TRAP_HTTP_EXCEPTIONS",
        "TRAP_BAD_REQUEST_ERRORS",
        "USE_X_SENDFILE",
    ]

    for flask_config in FLASK_CONFIGS_ALLOWED:
        flask_config_value = getenv_as_bool(
            f"{FLASK_CONFIG_IDENTIFIER}{flask_config}", default=None
        )
        if flask_config_value is None:
            continue
        app.config[flask_config] = flask_config_value
    logger.info(f"App Config:  {app.config}")


# ----------------------------
# GRPC
# ----------------------------


class SeldonModelGRPC:
    def __init__(self, user_model, seldon_metrics):
        self.user_model = user_model
        self.seldon_metrics = seldon_metrics

        self.metadata_data = seldon_core.seldon_methods.init_metadata(user_model)

    def Predict(self, request_grpc, context):
        return seldon_core.seldon_methods.predict(
            self.user_model, request_grpc, self.seldon_metrics
        )

    def SendFeedback(self, feedback_grpc, context):
        return seldon_core.seldon_methods.send_feedback(
            self.user_model, feedback_grpc, PRED_UNIT_ID, self.seldon_metrics
        )

    def TransformInput(self, request_grpc, context):
        return seldon_core.seldon_methods.transform_input(
            self.user_model, request_grpc, self.seldon_metrics
        )

    def TransformOutput(self, request_grpc, context):
        return seldon_core.seldon_methods.transform_output(
            self.user_model, request_grpc, self.seldon_metrics
        )

    def Route(self, request_grpc, context):
        return seldon_core.seldon_methods.route(
            self.user_model, request_grpc, self.seldon_metrics
        )

    def Aggregate(self, request_grpc, context):
        return seldon_core.seldon_methods.aggregate(
            self.user_model, request_grpc, self.seldon_metrics
        )

    def Metadata(self, request_grpc, context):
        """Metadata method of rpc Model service"""
        return json_to_seldon_model_metadata(self.metadata_data)

    def ModelMetadata(self, request_grpc, context):
        """ModelMetadata method of rpc Seldon service"""
        return json_to_seldon_model_metadata(self.metadata_data)

    def GraphMetadata(self, request_grpc, context):
        """GraphMetadata method of rpc Seldon service"""
        raise NotImplementedError("GraphMetadata not available on the Model level.")


def get_grpc_server(
    user_model, seldon_metrics, annotations={}, trace_interceptor=None, num_threads=1
):
    seldon_model = SeldonModelGRPC(user_model, seldon_metrics)
    options = [
        ("grpc.so_reuseport", 1),
    ]
    if ANNOTATION_GRPC_MAX_MSG_SIZE in annotations:
        max_msg = int(annotations[ANNOTATION_GRPC_MAX_MSG_SIZE])
        logger.info("Setting grpc max message and receive length to %d", max_msg)
        options.append(("grpc.max_message_length", max_msg))
        options.append(("grpc.max_send_message_length", max_msg))
        options.append(("grpc.max_receive_message_length", max_msg))

    server = grpc.server(
        futures.ThreadPoolExecutor(max_workers=num_threads), options=options
    )

    if trace_interceptor:
        from grpc_opentracing.grpcext import intercept_server

        server = intercept_server(server, trace_interceptor)

    prediction_pb2_grpc.add_GenericServicer_to_server(seldon_model, server)
    prediction_pb2_grpc.add_ModelServicer_to_server(seldon_model, server)
    prediction_pb2_grpc.add_TransformerServicer_to_server(seldon_model, server)
    prediction_pb2_grpc.add_OutputTransformerServicer_to_server(seldon_model, server)
    prediction_pb2_grpc.add_CombinerServicer_to_server(seldon_model, server)
    prediction_pb2_grpc.add_RouterServicer_to_server(seldon_model, server)
    prediction_pb2_grpc.add_SeldonServicer_to_server(seldon_model, server)

    SERVICE_NAMES = (
        prediction_pb2.DESCRIPTOR.services_by_name["Generic"].full_name,
        prediction_pb2.DESCRIPTOR.services_by_name["Model"].full_name,
        prediction_pb2.DESCRIPTOR.services_by_name["Router"].full_name,
        prediction_pb2.DESCRIPTOR.services_by_name["Transformer"].full_name,
        prediction_pb2.DESCRIPTOR.services_by_name["OutputTransformer"].full_name,
        prediction_pb2.DESCRIPTOR.services_by_name["Combiner"].full_name,
        prediction_pb2.DESCRIPTOR.services_by_name["Seldon"].full_name,
        reflection.SERVICE_NAME,
    )
    reflection.enable_server_reflection(SERVICE_NAMES, server)

    return server



# Copyright 2019 kubeflow.org.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#
# Copied from kfserving project as starter.
#

import pytest

import seldon_core

STORAGE_MODULE = "seldon_core.storage"


def test_storage_local_path():
    abs_path = "file:///"
    relative_path = "file://."
    assert seldon_core.Storage.download(abs_path) == abs_path.replace("file://", "", 1)
    assert seldon_core.Storage.download(relative_path) == relative_path.replace(
        "file://", "", 1
    )


@pytest.mark.parametrize(
    "uri",
    [
        "file:///not/existing/path",
        "s3://foo/bar",
        "gs://foo/bar",
        "https://accountname.blob.core.windows.net/container/some/blob/",
    ],
)
def test_invalid_storage_uri(uri):
    with pytest.raises(Exception):
        seldon_core.Storage.download(uri)


def test_no_prefix_local_path():
    abs_path = "/"
    relative_path = "."
    assert seldon_core.Storage.download(abs_path) == abs_path
    assert seldon_core.Storage.download(relative_path) == relative_path



import logging
import os

import pytest

import seldon_core
from seldon_core.metrics import SeldonMetrics
from seldon_core.wrapper import get_rest_microservice

from .helpers import MicroserviceWrapper, UserObject

logging.basicConfig(level=logging.DEBUG)

RESOURCES_PATH = os.path.join(os.path.dirname(__file__), "resources")


@pytest.fixture(params=[True, False])
def client_gets_metrics(monkeypatch, request):
    value = request.param
    monkeypatch.setattr(
        seldon_core.user_model, "INCLUDE_METRICS_IN_CLIENT_RESPONSE", value
    )
    monkeypatch.setattr(
        seldon_core.seldon_methods, "INCLUDE_METRICS_IN_CLIENT_RESPONSE", value
    )
    return value


@pytest.fixture
def microservice(request):
    # Make it compatible for both direct and indirect usage
    opts = {}
    if hasattr(request, "param"):
        opts = request.param

    # Extract opts from request' param
    app_name = opts.get("app_name", "model-template-app")
    app_location = opts.get("app_location", os.path.join(RESOURCES_PATH, app_name))
    envs = opts.get("envs", {})
    tracing = opts.get("tracing", False)

    wrapper = MicroserviceWrapper(app_location=app_location, envs=envs, tracing=tracing)

    with wrapper:
        yield wrapper


@pytest.fixture
def app():
    user_object = UserObject()
    seldon_metrics = SeldonMetrics()
    return get_rest_microservice(user_object, seldon_metrics)
